""" Functions for validating spliced PSMs.
"""
import itertools
import re

from Bio import SeqIO
import numpy as np
import pandas as pd
from scipy.stats import spearmanr

from inspire.constants import (
    CHARGE_KEY,
    KNOWN_PTM_WEIGHTS,
    MINIMAL_FEATURE_SET,
    PROSIT_IONS_KEY,
    PROTON,
    RESIDUE_WEIGHTS,
    SCAN_KEY,
    SPEARMAN_KEY,
    SOURCE_KEY,
)
from inspire.input.msp import msp_to_df
from inspire.predict_spectra import predict_spectra
from inspire.spectral_features import (
    calculate_spectral_angle,
    get_ion_masses,
    get_matches,
)
from inspire.utils import fetch_scan_data



def get_sa(
        df_row,
        mz_accuracy,
        mz_units,
    ):
    """ Function to extract the ion intensities from the true spectra which match
    """
    sequence = df_row['pcpPeptide']
    prosit_preds = df_row[PROSIT_IONS_KEY]
    mod_seq = df_row['modified_sequence']


    if '(ox)' not in mod_seq and 'C' not in mod_seq:
        mods = None
    else:
        mods = '0.'
        while mod_seq:
            if len(mod_seq) > 1:
                if mod_seq[1] == '(':
                    mods += '1'
                    mod_seq = mod_seq[5:]
                    continue
            if mod_seq[0] != 'C':
                mods += '0'
            else:
                mods += '2'
            mod_seq = mod_seq[1:]
        mods += '.0'
    potential_ion_mzs, precursor_weight = get_ion_masses(
        sequence,
        {
            0: 0.0,
            1: KNOWN_PTM_WEIGHTS['Oxidation (M)'],
            2: KNOWN_PTM_WEIGHTS['Carbamidomethylation'],
        },
        modifications=mods
    )

    precursor_mz = (precursor_weight + (PROTON*df_row[CHARGE_KEY]))/df_row[CHARGE_KEY]

    (
        match_info, _, __,
        ___, ____
    ) = get_matches(
        potential_ion_mzs,
        prosit_preds,
        df_row['mzs'],
        precursor_mz,
        df_row['intensities'],
        mz_accuracy,
        df_row['charge'],
        mz_units,
    )

    matched_intensities = match_info['matched_intensities']
    ordered_prosit_intes = match_info['ordered_prosit_intes']

    matched_l2_norm = np.linalg.norm(matched_intensities, ord=2)

    if matched_l2_norm:
        normed_matched_intensities = matched_intensities/matched_l2_norm
    else:
        normed_matched_intensities = matched_intensities

    df_row['spectralAngle'] = calculate_spectral_angle(
        normed_matched_intensities,
        ordered_prosit_intes,
    )
    spearman_total = spearmanr(normed_matched_intensities, ordered_prosit_intes)[0]

    if not np.isnan(spearman_total):
        df_row[SPEARMAN_KEY] = spearman_total
    else:
        df_row[SPEARMAN_KEY] = 0.0

    return df_row



def get_cleavage_peptides(proteome_loc):
    """ Function to find all nonspliced peptides.
    """
    prot_sequences = list(SeqIO.parse(proteome_loc, 'fasta'))

    weights = []
    pcp_seqs = []
    for protein in prot_sequences:
        prot_seq = str(protein.seq)
        for pep_len in range(7, 31):
            for i in range(1 + len(prot_seq) - pep_len):
                pep = prot_seq[i:i+pep_len]
                pcp_seqs.append(pep)
                weights.append(
                    round(sum([RESIDUE_WEIGHTS[z] for z in pep]), 2)
                )

    return pd.DataFrame({
        'pcpPeptide': pd.Series(pcp_seqs),
        'mw': pd.Series(weights),
    })

def get_mod_matched_seqs(mod_seq, pcp_pep):
    """ Funciton to find all matched sequences including oxidation of methionine.
    """
    n_mod_ox = mod_seq.count('[+16.0]')
    if n_mod_ox == 0:
        return [pcp_pep]
    n_methionine = pcp_pep.count('M')
    if n_methionine < n_mod_ox:
        return []
    methionine_pos = [
        m.start() for m in re.finditer('M', pcp_pep)
    ]
    possible_oxidations = itertools.combinations(methionine_pos, n_mod_ox)

    possible_mod_seqs = []
    for possible_os in possible_oxidations:
        mod_str = pcp_pep
        for o_site in sorted(possible_os, reverse=True):
            mod_str = mod_str[:o_site+1] + '[+16.0]' + mod_str[o_site+1:]
        possible_mod_seqs.append(mod_str)
    return possible_mod_seqs

def find_competitors(final_df, config):
    """ Function to find isobaric nonspliced competitors to spliced peptides.
    """
    final_df = final_df[
        (final_df['accessionGroup'] == 'spliced')
    ]

    final_df['mw'] = final_df['peptide'].apply(
        lambda x : round(sum([RESIDUE_WEIGHTS[z] for z in x]), 2)
    )

    pcp_df = get_cleavage_peptides(config.proteome)
    pcp_df = pcp_df.drop_duplicates()
    pcp_df['label'] = 1

    merged_df = pd.merge(
        final_df[[
            'source',
            'scan',
            'peptide',
            'modifiedSequence',
            'charge',
            'retentionTime',
            'proteins',
            'mw'
        ]],
        pcp_df[['pcpPeptide', 'mw', 'label']],
        how='inner',
        on='mw'
    )
    if merged_df.shape[0] == 0:
        return merged_df
    merged_df['matched_pcps'] = merged_df.apply(
        lambda x : get_mod_matched_seqs(x['modifiedSequence'], x['pcpPeptide']),
        axis=1,
    )
    merged_df = merged_df.explode('matched_pcps')
    merged_df = merged_df[merged_df['matched_pcps'].apply(lambda x : isinstance(x, str))]
    merged_df = merged_df.sort_values('peptide')
    merged_df['modified_sequence'] = merged_df['matched_pcps'].apply(
        lambda x : x.replace('[+16.0]', '(ox)').replace('[+57.0]', '')
    )

    prosit_input_df = merged_df.rename(columns={'charge': 'precursor_charge'})
    prosit_input_df['collision_energy'] = config.collision_energy
    prosit_input_df = prosit_input_df[[
        'modified_sequence', 'precursor_charge', 'collision_energy'
    ]].drop_duplicates()
    prosit_input_df.to_csv(
        f'{config.output_folder}/validationInput.csv', index=False
    )
    predict_spectra(config, 'validation')
    return merged_df

def calculate_competitor_spectral_data(competitors_df, config):
    """ Function to calculate spectral angles for isobaric nonspliced competitors.
    """
    scan_df = fetch_scan_data(competitors_df, config, with_charge=False)
    scan_df = scan_df.to_pandas()
    competitors_df = pd.merge(
        competitors_df,
        scan_df,
        how='inner',
        on=[SOURCE_KEY, SCAN_KEY]
    )

    prosit_df = msp_to_df(
        f'{config.output_folder}/validationPredictions.msp', 'prosit', None
    )
    prosit_df = prosit_df.drop_duplicates(subset=['modified_sequence', 'charge'])
    competitors_df['modified_sequence'] = competitors_df['matched_pcps'].apply(
        lambda x : x.replace('[+16.0]', '(ox)').replace('[+57.0]', '')
    )
    competitors_df = pd.merge(
        competitors_df,
        prosit_df,
        how='inner',
        on=['modified_sequence', CHARGE_KEY]
    )

    competitors_df['peptide'] = competitors_df['pcpPeptide']
    competitors_df = competitors_df.apply(
        lambda x : get_sa(x, config.mz_accuracy, config.mz_units),
        axis=1
    )
    return competitors_df

def validate_spliced(config):
    """ Funciton to validate spliced PSMs against isobaric nonspliced competitors.
    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the experiment.
    """
    final_df = pd.read_csv(
        f'{config.output_folder}/finalAssignments.csv'
    )
    final_df = final_df[final_df['qValue'] < 0.01]
    if config.contaminant_data is not None:
        final_df = filter_contaminants(final_df, config)
    competitors_df = find_competitors(final_df, config)

    if not competitors_df.shape[0]:
        final_df.to_csv(
            f'{config.output_folder}/filtered_finalAssignments.csv',
            index=False,
        )

    competitors_df = calculate_competitor_spectral_data(competitors_df, config)

    rt_ints = {}
    rt_coefs = {}
    for raw_file in competitors_df[SOURCE_KEY].unique().tolist():
        rt_df = pd.read_csv(
            f'{config.output_folder}/rt_fit_{raw_file}.csv'
        )
        rt_ints[raw_file] = rt_df['intercepts'].mean()

        rt_coefs[raw_file] = rt_df['coefficents'].mean()

    competitors_df['predRT'] = competitors_df.apply(
        lambda x : rt_ints[x['source']] + (rt_coefs[x['source']]*x['iRT']),
        axis=1,
    )
    competitors_df['base_deltaRT'] = competitors_df.apply(
        lambda x : abs(
            (x['retentionTime'] - x['predRT'])
        ),
        axis=1,
    )
    competitors_df['deltaRT'] = competitors_df.apply(
        lambda x : abs(
            abs(
                x['base_deltaRT']/rt_coefs.get(x['source'], 1.0)
            ) if rt_coefs.get(x['source'], 1.0) > 0 else x['base_deltaRT']
        ),
        axis=1
    )

    competitors_df['accession_spliced'] = 0
    competitors_df['accession_nonspliced'] = 1
    competitors_df.drop(
        ['matched_pcps','intensities','mzs','prositIons','iRT'], axis=1,
    ).sort_values(by='spectralAngle', ascending=False).to_csv(
        f'{config.output_folder}/cleavageCompetitors.csv',
        index=False,
    )

    weights_df = pd.read_csv(
        f'{config.output_folder}/final.percolatorSeparate.weights.csv',
        skiprows=lambda x : x not in [0, 2, 5, 8],
        sep='\t',
    )

    scores = np.matmul(
        competitors_df[MINIMAL_FEATURE_SET].values,
        weights_df[MINIMAL_FEATURE_SET].transpose().values
    )

    bias_terms = weights_df['m0'].values
    scores += bias_terms.transpose()

    final_scores = np.average(scores, axis=1)

    competitors_df['compScore'] = final_scores
    competitors_df = competitors_df.sort_values(by='compScore', ascending=False)
    competitors_df = competitors_df.drop_duplicates(subset=[SOURCE_KEY, SCAN_KEY])

    competitors_df = competitors_df.reset_index(drop=True)
    competitors_df['group'] = competitors_df.index // 10
    competitors_df['index'] = competitors_df.index % 10

    final_df = pd.read_csv(
        f'{config.output_folder}/finalAssignments.csv'
    )
    for possible_col in ['pcpPeptide',' modified_sequence']:
        if possible_col in final_df.columns:
            final_df = final_df.drop(possible_col, axis=1)

    final_df = pd.merge(
        final_df,
        competitors_df[[
            SOURCE_KEY, SCAN_KEY, 'pcpPeptide', 'modified_sequence', 'compScore', 'group', 'index'
        ]],
        how='left',
        on=[SOURCE_KEY, SCAN_KEY]
    )
    final_df[final_df['pcpPeptide'].apply(lambda x : isinstance(x, str))].to_csv(
        f'{config.output_folder}/competitorPsp.csv'
    )
    for raw_file in final_df[SOURCE_KEY].unique().tolist():
        rt_df = pd.read_csv(
            f'{config.output_folder}/rt_fit_{raw_file}.csv'
        )
        rt_ints[raw_file] = rt_df['intercepts'].mean()
        rt_coefs[raw_file] = rt_df['coefficents'].mean()

    final_df['deltaRT'] = final_df.apply(
        lambda x : abs(
            x['deltaRT']/rt_coefs.get(
                x['source'], 1.0
            ) if rt_coefs.get(x['source'], 1.0) > 0 else x['deltaRT']
        ),
        axis=1,
    )

    final_df['accession_nonspliced'] = final_df['accessionGroup'].apply(
        lambda x : 1 if x == 'nonspliced' else 0
    )
    final_df['accession_spliced'] = final_df['accessionGroup'].apply(
        lambda x : 1 if x == 'spliced' else 0
    )
    re_scores = np.matmul(
        final_df[MINIMAL_FEATURE_SET].values,
        weights_df[MINIMAL_FEATURE_SET].transpose().values
    )

    re_scores += bias_terms
    final_df['reScore1'] = re_scores[:,0]
    final_df['reScore2'] = re_scores[:,1]
    final_df['reScore3'] = re_scores[:,2]
    final_df['reScore'] = np.average(re_scores, axis=1)

    final_df['compScore']  = final_df['compScore'].fillna(-100_000)

    final_df[final_df.apply(
        lambda x : x['compScore'] is not None and x['reScore'] < x['compScore'],
        axis=1,
    )].to_csv(f'{config.output_folder}/knockout.csv', index=False)

    final_df = final_df[final_df.apply(
        lambda x : x['compScore'] is None or x['reScore'] > x['compScore'],
        axis=1,
    )]

    final_df = final_df.drop(['compScore'], axis=1)
    final_df = final_df.drop(
        [
            'group',
            'index',
            'accession_nonspliced',
            'accession_spliced',
            'reScore1',
            'reScore2',
            'reScore3',
            'reScore',
        ],
        axis=1,
    )

    final_df.to_csv(
        f'{config.output_folder}/filtered_finalAssignments.csv',
        index=False,
    )


def _separate_psm_id(df_row):
    """ Function to separate contaminant assignments.
    """
    split_psm_id = df_row['specID'].split('_')
    source = '_'.join(split_psm_id[:-2])
    scan = int(split_psm_id[-2])
    mod_pep = split_psm_id[-1]
    df_row['source'] = source
    df_row['scan'] = scan
    df_row['contModPep'] = mod_pep
    df_row['contPeptide'] = mod_pep.replace('[+1.0]', '').replace('[+16.0]', '').replace('[+57.0]', '')
    return df_row

def cont_filter(df_row):
    if not isinstance(df_row['contPeptide'], str):
        return True
    if df_row['peptide'].replace('I', 'L') == df_row['contPeptide'].replace('I', 'L'):
        return True
    if df_row['spectralAngle'] < df_row['contamSA']:
        return False
    return True


def filter_contaminants(final_df, config):
    final_df = final_df[
        (final_df['qValue'] < 0.01)
    ]

    contam_df = pd.read_csv(
        config.contaminant_data, sep='\t'
    )

    contam_df = contam_df[['specID', 'deltaRT', 'spectralAngle']]
    contam_df = contam_df[
        (contam_df['spectralAngle'] > final_df['spectralAngle'].min()*0.9)
    ]
    contam_df = contam_df.apply(_separate_psm_id, axis=1)
    contam_df = contam_df.sort_values(by='spectralAngle', ascending=False)
    contam_df = contam_df.drop_duplicates(subset=['source', 'scan'])
    contam_df = contam_df.drop(['specID'], axis=1)
    contam_df = contam_df.rename(columns={
        'spectralAngle': 'contamSA',
        'deltaRT': 'contamDeltaRT',
    })
    contam_df = contam_df[['source', 'scan', 'contPeptide', 'contamSA']]

    final_df = pd.merge(
        final_df,
        contam_df,
        how='left',
        on=['source', 'scan'],
    )

    filtered_final_df = final_df[final_df.apply(cont_filter, axis=1)]
    filtered_final_df = filtered_final_df.drop(['contamSA', 'contPeptide'], axis=1)

    return filtered_final_df



""" Functions for calculating difference between predicted and detected
    retention times.
"""
import os

import pandas as pd
import polars as pl
from pyteomics import achrom
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import KFold

from inspire.constants import (
    ACCESSION_STRATUM_KEY,
    LABEL_KEY,
    PEPTIDE_KEY,
    RT_KEY,
    SPECTRAL_ANGLE_KEY,
)

def _add_achrom_rt_preds(train_df, test_df):
    """ Function to fit and predict achrom's predictor.

    Parameters
    ----------
    train_df : pd.DataFrame
        A DataFrame on which to fit the retention time predictor.
    test_df : pd.DataFrame
        A DataFrame on which to predict retention time.

    Returns
    -------
    test_df : pd.DataFrame
        The input test_df with predRT as a new column.
    """
    retention_time_model = achrom.get_RCs_vary_lcp(
        train_df[PEPTIDE_KEY],
        train_df['retentionTime'],
        lcp_range=(-1.0, 1.0),
        term_aa=False,
        rcond=None
    )

    if 'C' not in retention_time_model['aa']:
        retention_time_model['aa']['C'] = retention_time_model['aa']['P']*0.956
    if 'W' not in retention_time_model['aa']:
        retention_time_model['aa']['W'] = retention_time_model['aa']['P']*0.956

    pred_rts = test_df[PEPTIDE_KEY].apply(
        lambda x :  achrom.calculate_RT(x, retention_time_model, raise_no_mod=False)
    )
    return pred_rts

def add_delta_irt(combined_df, config, scan_file):
    """ Function to calculate difference between predicted and observed retention
        time for each PSM.

    Parameters
    ----------
    combined_df : pd.DataFrame
        A DataFrame of PSMs.

    Returns
    -------
    combined_df : pd.DataFrame
        The DataFrame updated with a deltaRT column.
    """
    if combined_df[RT_KEY].n_unique() <= 1:
        combined_df['deltaRT'] = 0
        return combined_df

    kfold = KFold(n_splits=10, shuffle=True, random_state=42)
    combined_df_list = []
    coefficents = []
    intercepts = []

    if 'iRT' in combined_df.columns:
        combined_df_irt_null_count = combined_df.filter(
            pl.col('iRT').is_null()
        ).shape[0]
    else:
        combined_df_irt_null_count = 1

    if config.rt_fit_loc is not None:
        if os.path.exists(f'{config.rt_fit_loc}/rt_fit_{scan_file}.csv'):
            rt_df = pd.read_csv(f'{config.rt_fit_loc}/rt_fit_{scan_file}.csv')
        else:
            rt_dfs = [x for x in os.listdir(config.rt_fit_loc) if x.startswith('rt_fit')]
            rt_df = pd.read_csv(f'{config.rt_fit_loc}/{rt_dfs[0]}')

        coef = rt_df['coefficents'].mean()
        intercept = rt_df['intercepts'].mean()
        combined_df = combined_df.with_columns(
            pl.col('iRT').apply(lambda x : (x*coef)+ intercept).alias('predRT')
        )
        combined_df = combined_df.with_columns(
            (pl.col('predRT') - pl.col('retentionTime')).truediv(coef).abs().alias('deltaRT')
        )
        return combined_df

    try:
        for train, test in kfold.split(combined_df):
            train_df = combined_df[train, :]
            test_df = combined_df[test, :]

            if ACCESSION_STRATUM_KEY in train_df.columns:
                train_df = train_df.filter(
                    pl.col(ACCESSION_STRATUM_KEY).eq(0)
                )

            if LABEL_KEY in train_df.columns:
                train_df = train_df.filter(
                    pl.col(LABEL_KEY).eq(1)
                )

            top_spec_angle_cut = train_df[SPECTRAL_ANGLE_KEY].quantile(0.9)
            train_df = train_df.filter(
                (pl.col(SPECTRAL_ANGLE_KEY) > top_spec_angle_cut)
            )

            if combined_df_irt_null_count > 0:
                test_df['predRT'] = _add_achrom_rt_preds(train_df, test_df)
            else:
                reg = LinearRegression().fit(
                    train_df.select('iRT').to_numpy(),
                    train_df[RT_KEY].to_numpy(),
                )
                coefficents.append(reg.coef_[0])
                intercepts.append(reg.intercept_)
                pred_rt = pl.Series(
                    reg.predict(test_df.select('iRT').to_numpy())
                )
                test_df = test_df.with_columns(
                    pred_rt.alias('predRT')
                )

            test_df = test_df.with_columns(
                (pl.col('predRT') - pl.col('retentionTime')).abs().alias('deltaRT')
            )
            combined_df_list.append(test_df)
    except ValueError:
        try:
            # Simplest way to avoid errors on tiny files.
            train_df = train_df.filter(pl.col(LABEL_KEY).eq(1))
            reg = LinearRegression().fit(
                train_df.select('iRT').to_numpy(),
                train_df[RT_KEY].to_numpy(),
            )
            coefficents.append(reg.coef_[0])
            intercepts.append(reg.intercept_)
            pred_rt = pl.Series(
                reg.predict(combined_df.select('iRT').to_numpy())
            )
            combined_df = combined_df.with_columns(
                pred_rt.alias('predRT')
            )
            combined_df = combined_df.with_columns(
                (pl.col('predRT') - pl.col('retentionTime')).abs().alias('deltaRT')
            )
        except ValueError:
            reg = LinearRegression().fit(
                combined_df.select('iRT').to_numpy(),
                combined_df[RT_KEY].to_numpy(),
            )
            coefficents.append(reg.coef_[0])
            intercepts.append(reg.intercept_)
            pred_rt = pl.Series(
                reg.predict(combined_df.select('iRT').to_numpy())
            )
            combined_df = combined_df.with_columns(
                pred_rt.alias('predRT')
            )
            combined_df = combined_df.with_columns(
                (pl.col('predRT') - pl.col('retentionTime')).abs().alias('deltaRT')
            )


    rt_df = pd.DataFrame({
        'coefficents': coefficents,
        'intercepts': intercepts,
    })
    if scan_file is not None:
        rt_df.to_csv(f'{config.output_folder}/rt_fit_{scan_file}.csv', index=False)

    return pl.concat(combined_df_list)



""" Functions for providing a report at the end of inSPIRE rescoring.
"""
import pickle
import pandas as pd
import polars as pl
import yaml

from inspire.constants import (
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    ENDC_TEXT,
    ENGINE_SCORE_KEY,
    FINAL_Q_VALUE_KEY,
    FINAL_SCORE_KEY,
    IN_ACCESSION_KEY,
    MASS_DIFF_KEY,
    OKCYAN_TEXT,
    OUT_PSM_ID_KEY,
    OUT_Q_KEY,
    OUT_SCORE_KEY,
    PEPTIDE_KEY,
    PREFIX_KEYS,
    PSM_ID_KEY,
    SUFFIX_KEYS,
    PRED_ACCESSION_KEY,
    PRED_PEPTIDE_KEY,
    SEQ_LEN_KEY,
    TRUE_PEPTIDE_KEY,
)
from inspire.figures import (
    create_binders_fig,
    create_psms_fig,
    create_violin_fig,
    create_weights_fig,
)
from inspire.html_template import create_html_report
from inspire.input.mhcpan import read_mhcpan_output
from inspire.rescore import apply_rescoring

NON_SPECTRAL_FEATURES = [
    ENGINE_SCORE_KEY,
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    MASS_DIFF_KEY,
    SEQ_LEN_KEY,
    'nVarMods',
    'avgResidueMass',
]
N_PR_STEPS = 200

def combine_results(output_folder, mhc_pan_df, input_file):
    """ Function to add NetMHCpan predictions to inSPIRE PSMs.

    Parameters
    ----------
    output_folder : str
        The folder where all inSPIRE output is written.
    mhc_pan_df : pd.DataFrame
        A DataFrame of NetMHCpan predictions.
    input_file : str
        The input file of rescored psms.

    Returns
    -------
    final_assignments_df : pd.DataFrame
        Combined PSMs and NetMHCpan predictions.
    """
    if 'non_spectral' in input_file:
        final_assignments_df = pd.read_csv(f'{output_folder}/{input_file}', sep='\t')
        final_assignments_df[PEPTIDE_KEY] = final_assignments_df[PEPTIDE_KEY].apply(
            lambda x : x.split('.')[1] if '.' in x else x
        )

        if 'mokapot' in input_file:
            score_key = OUT_SCORE_KEY['mokapot']
        else:
            score_key = OUT_SCORE_KEY['percolator']
    else:
        final_assignments_df = pd.read_csv(f'{output_folder}/{input_file}')
        score_key = FINAL_SCORE_KEY

    mhc_pan_df = mhc_pan_df.rename(columns={'Peptide': PEPTIDE_KEY})
    mhc_pan_df = mhc_pan_df[[PEPTIDE_KEY, '%Rank_BA', 'BindLevel']]
    if 'BindLevel' in final_assignments_df.columns:
        final_assignments_df = final_assignments_df.drop('BindLevel', axis=1)
    final_assignments_df = pd.merge(
        final_assignments_df,
        mhc_pan_df,
        on=PEPTIDE_KEY,
        how='left',
    )

    final_assignments_df = final_assignments_df.sort_values(
        by=score_key,
        ascending=False
    )

    return final_assignments_df

def apply_non_spectral_percolator(
        output_folder,
        fdr,
        rescore_method,
        rescore_command,
        proteome,
        use_score_only=False
    ):
    """ Function to apply percolator without spectral features as a comparison with
        full inSPIRE.

    Parameters
    ----------
    output_folder : str
        The folder where all inSPIRE output is written.
    fdr : float
        The false discovery rate we are aiming.
    rescore_method : str
        The method to be used for rescoring.
    proteome : str or None
        Path to proteome file is protein inference to be performed.
    use_score_only : bool
        Flag indicating whether or not to use the engine score only.

    Returns
    -------
    non_spectral_psm_df : pd.DataFrame
        A DataFrame of results from Percolator trained without spectral features.
    """
    all_features_df = pl.from_pandas(
        pd.read_csv(f'{output_folder}/input_all_features.tab', sep='\t')
    )
    all_features_df = all_features_df.filter(
        pl.col(IN_ACCESSION_KEY[rescore_method]).ne('deNovo')
    )

    prefix_keys = PREFIX_KEYS[rescore_method]
    psm_id_key = PSM_ID_KEY[rescore_method]

    if proteome is not None:
        all_features_df = all_features_df.with_columns(
            (pl.lit('-.') + pl.col(PEPTIDE_KEY) + pl.lit('.-')).alias(PEPTIDE_KEY)
        )

    if use_score_only:
        non_spectral_df = all_features_df.select(
            prefix_keys +
            [ENGINE_SCORE_KEY] +
            SUFFIX_KEYS
        )
    else:
        non_spectral_df = all_features_df.select(
            prefix_keys + NON_SPECTRAL_FEATURES + SUFFIX_KEYS[rescore_method]
        )

    non_spectral_df.write_csv(
        f'{output_folder}/non_spectral_perc_input.tab',
        separator='\t',
    )

    non_spectral_psm_df, _ = apply_rescoring(
        output_folder,
        'non_spectral_perc_input.tab',
        fdr,
        rescore_method,
        'non_spectral',
        rescore_command,
        proteome,
    )
    non_spectral_psm_df = non_spectral_psm_df.rename(
        {OUT_PSM_ID_KEY[rescore_method]: psm_id_key}
    )

    non_spectral_psm_df = non_spectral_psm_df.join(
        non_spectral_df[[psm_id_key, ENGINE_SCORE_KEY]],
        how='inner',
        on=psm_id_key,
    )

    return non_spectral_psm_df

def get_stats_at_cut_off(assignment_df, binders_df, q_cut, q_val_key):
    """ Function to calculate statistics on the PSMs above a given cut off.

    Parameters
    ----------
    assignment_df : pd.DataFrame
        A DataFrame of PSM assignments.
    binders_df : pd.DataFrame
        A DataFrame of the predicted MHC binding of the PSMs.
    q_cut : float
        The q-value cut off to apply.
    q_val_key : str
        The name of the column containing q-values.

    Returns
    -------
    psm_count : int
        The number of psms above the threshold.
    binders_count : int
        The number of predicted binders above the threshold.
    pct_binders : float
        The percentage of predicted binders among qualifying PSMs.
    """
    psm_count = assignment_df.filter(pl.col(q_val_key).lt(q_cut)).shape[0]
    if binders_df is not None:
        binders_cut_df = binders_df[binders_df[q_val_key] <= q_cut]
        binders_divisor = binders_cut_df.shape[0]
        if binders_divisor > 0:
            binders_count = binders_cut_df[
                (binders_cut_df['BindLevel'] == '<=WB') | (binders_cut_df['BindLevel'] == '<=SB')
            ].shape[0]
            pct_binders = 100*binders_count/binders_divisor
        else:
            binders_count = 0
            pct_binders = 0
    else:
        binders_count = None
        pct_binders = None
    return psm_count, binders_count, pct_binders

def calculate_fdr_cut_offs(
        assignment_df,
        non_spectral_df,
        se_q_cut_key,
        binders_df=None,
        ns_binders_df=None
    ):
    """ Function calculate statistics on PSMs at different false discovery rates.

    Parameters
    ----------
    assignment_df : pd.DataFrame
        A DataFrame of inSPIRE final assignments.
    non_spectral_df : pd.DataFrame
        A DataFrame of assignments from Percolator without spectral features.
    rescore_method : str
        Indictor on whether mokapot or percolator is being used.
    binders_df : pd.DataFrame or None
        A DataFrame of NetMHCpan binders for inSPIRE assignments.
    ns_binders_df : pd.DataFrame or None
        A DataFrame of NetMHCpan binders for Percolator assignments without spectral features.

    Returns
    -------
    fdrs_df : pd.DataFrame
        A DataFrame of statistics on PSMs at various FDR thresholds.
    """
    q_cut_offs = [0.01*i for i in range(1, 11)]
    n_spire_psms = []
    n_search_engine_psms = []
    n_spire_binders = []
    n_search_engine_binders = []
    pct_spire_binders = []
    pct_search_engine_binders = []

    for q_cut in q_cut_offs:
        car_stats = get_stats_at_cut_off(
            assignment_df,
            binders_df,
            q_cut,
            FINAL_Q_VALUE_KEY,
        )
        n_spire_psms.append(car_stats[0])
        n_spire_binders.append(car_stats[1])
        pct_spire_binders.append(car_stats[2])

        se_stats = get_stats_at_cut_off(
            non_spectral_df,
            ns_binders_df,
            q_cut,
            se_q_cut_key,
        )
        n_search_engine_psms.append(se_stats[0])
        n_search_engine_binders.append(se_stats[1])
        pct_search_engine_binders.append(se_stats[2])

    return pd.DataFrame(
        {
            'FDR': q_cut_offs,
            'nSpirePsms': n_spire_psms,
            'nSpireBinders': n_spire_binders,
            'SpirePercentageBinders': pct_spire_binders,
            'nSearchEnginePsms': n_search_engine_psms,
            'nSearchEngineBinders': n_search_engine_binders,
            'searchEnginePercentageBinders': pct_search_engine_binders,
        }
    )

def _get_counts(all_df, score_cut_off, score_key, acc_grp):
    """ Function to get the predicted and correct counts above a scoring threshold.

    Parameters
    ----------
    all_df : pd.DataFrame
        A DataFrame of all psms.
    score_cut_off : float
        A cut off on psms percolator score.
    score_key : str
        The column name of the percolator scores.
    acc_grp : str
        The accession group we are interested in.

    Returns
    -------
    pred_count : int
        The number of PSMs for the accession group above the threshold.
    correct_count : int
        The number of correct PSMs for the accession group above the threshold.
    """
    if acc_grp != 'discoverable':
        filtered_df = all_df[
            (all_df[score_key] >= score_cut_off) &
            (all_df[PRED_ACCESSION_KEY] == acc_grp)
        ]
    else:
        filtered_df = all_df[
            all_df[score_key] >= score_cut_off
        ]

    pred_count = filtered_df.shape[0]

    if score_key.startswith('prosit'):
        pep_key = 'prositOverPeptide'
    else:
        pep_key = PRED_PEPTIDE_KEY

    correct_count = filtered_df[filtered_df.apply(
        lambda x : x[TRUE_PEPTIDE_KEY].replace('I', 'L') == x[pep_key].replace('I', 'L'),
        axis=1
    )].shape[0]

    return pred_count, correct_count


def generate_report(config):
    """ Function for creating a report at the end of ininspire execution.

    Parameters
    ----------
    config : inspire.config.Config
        The settings for the whole pipeline.
    """
    figures = {}
    figures['table'], most_pos_feats, most_neg_feats = create_weights_table(
        config.output_folder,
        config.rescore_method,
    )
    figures['violin_fig'] = calculate_distributions(
        config,
        most_pos_feats,
        most_neg_feats
    )

    if config.rescore_method == 'mokapot':
        non_spectral_fdr = 0.05
    else:
        non_spectral_fdr = config.fdr

    if config.infer_proteins:
        proteome = config.proteome
    else:
        proteome = None

    non_spectral_df = apply_non_spectral_percolator(
        config.output_folder,
        non_spectral_fdr,
        config.rescore_method,
        config.rescore_command,
        proteome,
    )

    assignment_df = pl.read_csv(f'{config.output_folder}/finalPsmAssignments.csv')

    if config.use_binding_affinity in ['asFeature', 'asValidation']:
        mhcpan_df = read_mhcpan_output(f'{config.output_folder}/mhcpan')
        mhcpan_df = mhcpan_df.to_pandas()
        binders_df = combine_results(config.output_folder, mhcpan_df, 'finalPsmAssignments.csv')
        ns_binders_df = combine_results(
            config.output_folder,
            mhcpan_df,
            f'non_spectral.{config.rescore_method}.psms.txt'
        )
        binders_df.to_csv(f'{config.output_folder}/binderfinalPsmAssignments.csv')

        fdrs_df = calculate_fdr_cut_offs(
            assignment_df,
            non_spectral_df,
            OUT_Q_KEY[config.rescore_method],
            binders_df,
            ns_binders_df
        )
    else:
        fdrs_df = calculate_fdr_cut_offs(
            assignment_df,
            non_spectral_df,
            OUT_Q_KEY[config.rescore_method],
        )

    if config.use_binding_affinity:
        figures['binders_fig'] = create_binders_fig(fdrs_df, config.search_engine)

    figures['psms_fig'] = create_psms_fig(fdrs_df, config.output_folder)
    fdrs_df.to_csv(f'{config.output_folder}/searchEngineFdrs.csv')
    print(
        OKCYAN_TEXT +
        '\tAll Figures Created.' +
        ENDC_TEXT
    )

    create_html_report(
        config,
        figures,
    )


def calculate_distributions(config, most_positive_features, most_negative_features):
    """ Function to produce violin plots.
    """
    out_score_key = OUT_SCORE_KEY[config.rescore_method]
    psm_id_key = PSM_ID_KEY[config.rescore_method]

    out_filename = f'final.{config.rescore_method}.psms.txt'
    input_df = pd.read_csv(f'{config.output_folder}/final_input.tab', sep='\t')
    psms_df = pd.read_csv(f'{config.output_folder}/{out_filename}', sep='\t').rename(
        columns={OUT_PSM_ID_KEY[config.rescore_method]: psm_id_key}
    )

    psms_df['Status'] = psms_df[out_score_key].apply(
        lambda x : 'Accepted' if x >= 0 else 'Rejected'
    )

    combined_df = pd.merge(
        input_df,
        psms_df[[psm_id_key, 'Status', OUT_SCORE_KEY[config.rescore_method]]],
        how='inner',
        on=psm_id_key,
    )

    violin_plot = create_violin_fig(combined_df, most_positive_features, most_negative_features)

    return violin_plot

def add_mokapot_weights(output_folder, rescore_method, acc_idx):
    """ Function to create a Figure of weights table.

    Parameters
    ----------
    output_folder : str
        The folder where all inspire output is written.
    rescore_method : str
        Indicator of whether mokapot or percolator is in use.

    Results
    -------
    weights_df : pd.DataFrame
        A DataFrame of the weights for different features.
    """
    with open(f'{output_folder}/selectedFeatures.yaml', 'r', encoding='UTF-8') as stream:
        feature_names = yaml.safe_load(stream)

    if acc_idx is None:
        prefix = 'final'
    else:
        prefix = f'final_{acc_idx}'

    feat_weights = {feature: [] for feature in feature_names}
    feat_weights['intercept'] = []
    for fold_idx in range(1, 4):
        with open(
            f'{output_folder}/{prefix}.{rescore_method}.model_fold-{fold_idx}.pkl',
            'rb',
        ) as model_file:
            model = pickle.load(model_file)
        weights = model.estimator.coef_[0]
        for feat_idx, name in enumerate(feature_names):
            feat_weights[name].append(round(weights[feat_idx], 2))
        feat_weights['intercept'].append(
            round(model.estimator.intercept_[0], 2)
        )

    return pd.DataFrame(feat_weights)

def clean_percolator_weights(weights_path):
    """ Function to remove the comments which are inexplicably added to a csv
        file.
    """
    with open(weights_path, 'r', encoding='UTF-8') as file :
        lines = file.readlines()
        relevant_lines = []
        for line in lines:
            if not line.startswith('#'):
                relevant_lines.append(line)

    with open(weights_path, 'w', encoding='UTF-8') as file:
        for line in relevant_lines:
            file.write(line)

def create_weights_table(output_folder, rescore_method, acc_idx=None):
    """ Function to create a Figure of weights table.

    Parameters
    ----------
    output_folder : str
        The folder where all inspire output is written.
    rescore_method : str
        Indicator of whether mokapot or percolator is in use.

    Returns
    -------
    fig : str
        A Figure showing weights converted to html.
    most_pos_feats : list of str
        A list of the most positively weighted features.
    most_neg_feats : list of str
        A list of the most negatively weighted features.
    """
    if rescore_method == 'mokapot':
        weights_df = add_mokapot_weights(output_folder, rescore_method, acc_idx)
    else:
        if acc_idx is None:
            weights_path = f'{output_folder}/final.{rescore_method}.weights.csv'
        else:
            weights_path = f'{output_folder}/final_{acc_idx}.{rescore_method}.weights.csv'
        clean_percolator_weights(weights_path)

        weights_df = pd.read_csv(
            weights_path,
            skiprows=lambda x : x not in [0, 1, 4, 7],
            sep='\t',
        )
    transposed_df = weights_df.transpose()

    weights_df_means = weights_df.mean()
    weight_keys = [f'weightFold{i}' for i in range(1, 4)]
    transposed_df.columns = weight_keys
    transposed_df['feature'] = transposed_df.index
    transposed_df['averageWeight'] = weights_df_means
    transposed_df = transposed_df.sort_values(by='averageWeight', ascending=False)
    m0_df = transposed_df[
        (transposed_df.index == 'm0') | (transposed_df.index == 'intercept')
    ]
    transposed_df = transposed_df[
        (transposed_df.index != 'm0') & (transposed_df.index != 'intercept')
    ]
    transposed_df = pd.concat([transposed_df, m0_df])

    weights_and_feats = sorted([
        (ave_wt, feat) for ave_wt, feat in zip(
            transposed_df.averageWeight.tolist(),
            transposed_df.feature.tolist()
        ) if feat != 'm0' and feat != 'intercept' and '_' not in feat
    ])
    most_neg_feats = [x[1] for x in weights_and_feats[:3]]
    most_pos_feats = [x[1] for x in weights_and_feats[-3:]]
    most_pos_feats.reverse()

    fig = create_weights_fig(transposed_df)
    return fig, most_pos_feats, most_neg_feats



""" Functions for writing percolator/mokapot input using Prosit and other features.
"""
from math import log10
import multiprocessing as mp
from multiprocessing.managers import BaseManager
import os
from pathlib import Path

import numpy as np
import polars as pl
from xgboost import XGBRegressor

from inspire.basic_features import create_basic_features
from inspire.constants import(
    ACCESSION_STRATUM_KEY,
    ACCESSION_KEY,
    BASIC_FEATURES,
    CHARGE_KEY,
    ENDC_TEXT,
    FRAG_MZ_ERR_VAR_KEY,
    FRAG_MZ_ERR_MED_KEY,
    LABEL_KEY,
    MINIMAL_FEATURE_SET,
    OKCYAN_TEXT,
    IN_ACCESSION_KEY,
    PEPTIDE_KEY,
    PSM_ID_KEY,
    PEARSON_KEY,
    PERC_SCAN_ID,
    PTM_ID_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    PTM_WEIGHT_KEY,
    RT_KEY,
    SCAN_KEY,
    SOURCE_INDEX_KEY,
    SOURCE_KEY,
    WARNING_TEXT,
)
from inspire.input.mascot import MASCOT_PEP_QUERY_KEY
from inspire.input.mgf import process_mgf_file
from inspire.input.mhcpan import read_mhcpan_output
from inspire.input.msp import msp_to_df
from inspire.input.mzml import process_mzml_file
from inspire.input.search_results import generic_read_df
from inspire.prepare import create_prosit_mod_seq
from inspire.retention_time import add_delta_irt
from inspire.spectral_features import (
    SPECTRAL_FEATURES,
    DELTA_FEATURES,
    create_spectral_features,
)
from inspire.utils import (
    accession_informed_filter,
    check_bad_mods,
    fetch_proteome,
    get_ox_flag,
    modify_sequence_for_skyline,
    parallel_remap,
    remove_source_suffixes,
)

BATCH_SIZE = 500

def combine_spectral_data(search_df, scan_df, prosit_df, ox_flag, spectral_predictor):
    """ Function to combine DataFrame of PEAKS results with prosit predicted
        spectrum and the true eperimental spectrum.

    Parameters
    ----------
    peaks_df : pd.DataFrame
        The DataFrame of PEAKS search results.
    mzml_df : pd.DataFrame
        The DataFrame of experimental observed spectra.
    prosit_df : pd.DataFrame
        The DataFrame of prosit predicted spectra.

    Returns
    -------
    final_df : pd.DataFrame
        The DataFrame with all merged Data.
    """
    if spectral_predictor == 'prosit':
        search_df = search_df.with_columns(
            pl.struct([PEPTIDE_KEY, PTM_SEQ_KEY]).apply(
                lambda x : create_prosit_mod_seq(x[PEPTIDE_KEY], x[PTM_SEQ_KEY], ox_flag),
                skip_nulls=False,
            ).alias('modified_sequence')
        )
        prosit_df = prosit_df.unique(
            subset=['modified_sequence', CHARGE_KEY, 'collisionEnergy']
        )


    print(
        OKCYAN_TEXT +
        f'\t\t\t{search_df.shape[0]} in original search results.' +
        ENDC_TEXT
    )
    if spectral_predictor == 'prosit':
        df_with_prosit = search_df.join(
            prosit_df,
            how='inner',
            on=['modified_sequence', CHARGE_KEY]
        )
    elif spectral_predictor == 'ms2pip':
        df_with_prosit = search_df.join(
            prosit_df,
            how='inner',
            on=[PEPTIDE_KEY, PTM_SEQ_KEY, CHARGE_KEY]
        )

    print(
        OKCYAN_TEXT +
        f'\t\t\t{df_with_prosit.shape[0]} after combination with predicted spectra.'
        + ENDC_TEXT
    )

    scan_df = scan_df.with_columns(
        pl.col(SOURCE_KEY).apply(remove_source_suffixes).alias(SOURCE_KEY)
    )

    final_df = df_with_prosit.join(
        scan_df,
        how='inner',
        on=[SOURCE_KEY, SCAN_KEY]
    )

    print(
        OKCYAN_TEXT +
        f'\t\t\t{final_df.shape[0]} after combination with experimental spectra.' +
        ENDC_TEXT
    )

    return final_df

def filter_input_columns(combined_df, config, file_idx):
    """ Function to filter the final percolator/mokapot input DataFrame to the required
        features.

    Parameters
    ----------
    combined_df : pd.DataFrame
        The percolator/mokapot input DataFrame.
    config : inspire.config.Config
        The config object.

    Returns
    -------
    combined_df : pd.DataFrame
        The input DataFrame containing only the features required for percolator/mokapot.
    """
    psm_id_key = PSM_ID_KEY[config.rescore_method]
    if config.minimal_features:
        use_cols = [
            psm_id_key,
            LABEL_KEY,
            PERC_SCAN_ID,
        ] + BASIC_FEATURES + MINIMAL_FEATURE_SET
    else:
        use_cols = [
            psm_id_key,
            LABEL_KEY,
            PERC_SCAN_ID,
        ] + BASIC_FEATURES + SPECTRAL_FEATURES

        if config.delta_method != 'ignore':
            use_cols += DELTA_FEATURES

        use_cols += ['deltaRT']

    if config.use_binding_affinity == 'asFeature':
        use_cols += ['bindingAffinity']

    if config.combined_scans_file is not None:
        if config.source_files is not None:
            scan_files = [remove_source_suffixes(x) for x in config.source_files]
        else:
            scan_files = sorted(combined_df[SOURCE_KEY].unique().tolist())
        combined_df = combined_df.with_columns(
            pl.col(SOURCE_KEY).apply(scan_files.index).alias(SOURCE_INDEX_KEY)
        )
    else:
        combined_df = combined_df.with_columns(
            pl.lit(file_idx).alias(SOURCE_INDEX_KEY)
        )

    if isinstance(config.collision_energy, list):
        use_cols += ['collisionEnergy']

    use_cols += [SOURCE_INDEX_KEY]


    if config.use_accession_stratum:
        acc_cols = [ACCESSION_STRATUM_KEY]
        for a_idx in range(len(config.accession_hierarchy)):
            stratum_name = config.accession_hierarchy[int(a_idx)]
            combined_df = combined_df.with_columns(
                pl.col(ACCESSION_STRATUM_KEY).apply(
                    lambda x, val=a_idx: 1 if x == val else 0
                ).alias(f'accession_{stratum_name}')
            )
            acc_cols.append(f'accession_{stratum_name}')

        use_cols += sorted(acc_cols)

    use_cols += [PEPTIDE_KEY, IN_ACCESSION_KEY[config.rescore_method]]

    if config.remap_to_proteome:
        proteome = fetch_proteome(config.proteome, with_desc=False)
        combined_df = parallel_remap(
            combined_df,
            config.n_cores,
            proteome,
            IN_ACCESSION_KEY[config.rescore_method],
        )
        combined_df = combined_df.drop(ACCESSION_KEY)
    else:
        combined_df = combined_df.rename(
            {ACCESSION_KEY: IN_ACCESSION_KEY[config.rescore_method]}
        )

    combined_df = combined_df.sort(by='tempIndex')
    combined_df = combined_df.select(use_cols)

    return combined_df

def write_with_spectral_features(
        search_df,
        mods_df,
        config,
    ):
    """ Function to write the percolator/mokapot input DataFrame with spectral features.

    Parameters
    ----------
    search_df : pd.DataFrame
        The results from the original search DataFrame.
    mods_df : pd.DataFrame
        The DataFrame of ptms.
    config : inspire.config.Config
        The Config object.
    """
    if config.combined_scans_file is not None:
        scan_files = sorted([remove_source_suffixes(config.combined_scans_file)])
    else:
        scan_files = sorted(search_df[SOURCE_KEY].unique().to_list())


    max_scan = search_df[SCAN_KEY].max()
    for file_idx, scan_file in enumerate(scan_files):
        func_args = generate_function_arguments(
            search_df, mods_df, config, file_idx, scan_file
        )
        process_single_file(
            func_args, config, file_idx, scan_file, max_scan
        )

    print(
        OKCYAN_TEXT + '\t\t\tFull input DataFrame written to csv.' + ENDC_TEXT
    )


def add_perc_scan_id(combined_df, file_idx, max_scan):
    """ Function add a Percolator scan ID to the DataFrame that will be unique
        to scans across RAW files.

    Parameters
    ----------
    combined_df : pl.DataFrame
        DataFrame of PSMs with combined spectral information.
    config : inspire.config.Config
        Config object for the whole experiment.
    file_idx : int
        The index of the raw file being processed.
    max_scan : int
        The highest scan number across all files.

    Returns
    -------
    combined_df : pl.DataFrame
        Input DataFrame with scannr column added.
    """
    combined_df = combined_df.with_columns(
        pl.col(SCAN_KEY).apply(
            lambda x, f_id=file_idx : f_id * max_scan  + x
        ).alias(PERC_SCAN_ID)
    )
        
    return combined_df

def generate_function_arguments(
        search_df, mods_df, config, file_idx, scan_file
    ):
    """ Function to process all PSMs from a single mgf or mzML file.

    Parameters
    ----------
    search_df : pd.DataFrame
        A DataFrame of PSMs.
    mods_df : pd.DataFrame
        A small DataFrame detailing the PTMs present in the data.
    prosit_df : pd.DataFrame
        A DataFrame of the predicted spectra from Prosit.
    config : inspire.config.Config
        The config file for the experiment.
    file_idx : int
        The index of the file being processed.
    scan_file : str
        The name of the file being processed.
    """
    ox_flag = get_ox_flag(mods_df)
    if config.spectral_predictor == 'prosit':
        prosit_predictions = f'{config.output_folder}/prositPredictions.msp'
        prosit_df = msp_to_df(prosit_predictions, 'prosit', None)
    elif config.spectral_predictor == 'ms2pip':
        model = config.ms2pip_model
        ms2pip_predictions = f'{config.output_folder}/ms2pipInput_{model}_predictions.msp'
        prosit_df = msp_to_df(ms2pip_predictions, 'ms2pip', mods_df)

    print(
        OKCYAN_TEXT +
        f'\t\tProcessing scan file {file_idx}, {scan_file}.' +
        ENDC_TEXT
    )
    if config.combined_scans_file is not None:
        filtered_search_df = search_df
    else:
        filtered_search_df = search_df.filter(
            pl.col(SOURCE_KEY).eq(scan_file)
        )

    scans = filtered_search_df.select(SCAN_KEY).to_series().unique()

    if RT_KEY not in filtered_search_df.columns:
        with_rt = True
    elif filtered_search_df.select(RT_KEY).n_unique() <= 1:
        filtered_search_df = filtered_search_df.drop(RT_KEY)
        with_rt = True
    else:
        with_rt = False

    if config.scans_format == 'mzML':
        scan_df = process_mzml_file(
            f'{config.scans_folder}/{scan_file}.{config.scans_format}',
            set(scans.to_list()),
            with_retention_time=with_rt,
        )
    else:
        scan_df = process_mgf_file(
            f'{config.scans_folder}/{scan_file}.{config.scans_format}',
            set(scans.to_list()),
            config.scan_title_format,
            config.source_files,
            combined_source_file=config.combined_scans_file is not None,
            with_retention_time=with_rt,
            with_ms1=True,
        )

    scan_df = scan_df.unique(subset=[SOURCE_KEY, SCAN_KEY])
    combined_df = combine_spectral_data(
        filtered_search_df,
        scan_df,
        prosit_df,
        ox_flag,
        config.spectral_predictor,
    )

    print(
        OKCYAN_TEXT + '\t\t\tCombined DB Search, Spectral, and Prosit Data.' + ENDC_TEXT
    )

    if not combined_df.shape[0]:
        print(
            WARNING_TEXT +
            f'Warning. No matched scans found for source file {scan_file}' +
            ENDC_TEXT
        )
        return None

    n_batches = combined_df.shape[0]//BATCH_SIZE
    additional_psms = combined_df.shape[0]%BATCH_SIZE
    batch_values = []
    for batch_idx in range(n_batches):
        batch_values.extend([batch_idx]*BATCH_SIZE)
    batch_values.extend([n_batches]*additional_psms)

    combined_df = combined_df.with_columns(
        pl.Series(name='batch', values=batch_values)
    )
    combined_df_list = combined_df.partition_by('batch')
    func_args = []
    for idx, comb_df in enumerate(combined_df_list):
        comb_df.write_parquet(f'{config.output_folder}/temp_{idx}_in.parquet')
        func_args.append([mods_df, config, idx])

    return func_args


class ChildRegressor(XGBRegressor):
    """ Class to allow use of Prosit-delta predictor in shared memory.
    """
    def __init__(self):
        super().__init__()
        home = str(Path.home())
        self.load_model(f'{home}/inSPIRE_models/models/prosit_delta_v1.json')
        self.set_params(n_jobs=1)
        self.set_params(random_state=42)

    def best_score(self):
        return 1.0

    def coef_(self):
        return np.array([1.0])

    def feature_names_in_(self):
        return np.array([''])

    def intercept_(self):
        return np.array([1.0])

class CustomManager(BaseManager):
    # nothing
    pass

def process_single_file(
        func_args, config, file_idx, scan_file, max_scan
    ):
    """ Function to process PSMs from a single raw file in parallel.

    Parameters
    ----------
    func_args : list of tuples
        A list of the arugments to be passed to each parallel execution of
        the create_spectral_features function.
    config : inspire.config.Config
        The Config object for the whole experiment.
    file_idx : int
        The index of the file being processed.
    scan_file : str
        The name of the file being processed.
    max_scan : int
        The maximum scan value in the dataset (needed to ensure uniqueness
        in Percolator scan ID).
    """
    CustomManager.register('ChildRegressor', ChildRegressor)
    with CustomManager() as manager:
        if config.delta_method != 'ignore' and not config.minimal_features:
            model = manager.ChildRegressor()
            # Avoid unecessary parallelism when model applied to small sets.
            model.set_params(n_jobs=1)
        else:
            model = None

        for entry in func_args:
            entry.append(model)

        func_args = [tuple(arg_group) for arg_group in func_args]

        with mp.get_context('spawn').Pool(processes=config.n_cores) as pool:
            pool.starmap(create_spectral_features, func_args)

    results_dfs = [
        pl.read_parquet(
            f'{config.output_folder}/temp_{idx}_out.parquet'
        ) for idx in range(len(func_args))
    ]
    select_columns = results_dfs[0].columns

    combined_df = pl.concat([x.select(select_columns) for x in results_dfs])
    replace_feats = [
        FRAG_MZ_ERR_VAR_KEY,
        FRAG_MZ_ERR_MED_KEY,
        PEARSON_KEY,
    ]
    if config.delta_method != 'ignore' and not config.minimal_features:
        replace_feats += DELTA_FEATURES

    for feat in set(replace_feats):
        if feat in select_columns:
            feat_median = pl.median(combined_df.filter(
                pl.col(feat).gt(-1)
            )[feat])

            combined_df = combined_df.with_columns(
                pl.when(pl.col(feat).eq(-1)).then(feat_median)
                    .otherwise(pl.col(feat)).alias(feat)
            )

    for idx in range(len(func_args)):
        os.remove(f'{config.output_folder}/temp_{idx}_in.parquet')
        os.remove(f'{config.output_folder}/temp_{idx}_out.parquet')

    combined_df = combined_df.sort(by='spectralAngle', descending=True)
    if isinstance(config.collision_energy, list):
        combined_df = combined_df.unique(subset=['source', 'scan', 'peptide'])

    combined_df = combined_df.sort(by='tempIndex')
    combined_df = add_delta_irt(combined_df, config, scan_file)

    print(
        OKCYAN_TEXT + '\t\t\tCreated Spectral and Delta RT Features.' + ENDC_TEXT
    )
    combined_df = add_perc_scan_id(combined_df, file_idx, max_scan)

    combined_df = filter_input_columns(combined_df, config, file_idx)

    file_loc = _write_to_tab_file(combined_df, file_idx, config.output_folder)

    return file_loc

def _write_to_tab_file(combined_df, file_idx, output_folder):
    """ Function to write percolator input in tab format.

    Parameters
    ----------
    combined_df : pd.DataFrame
        A DataFrame of processed search results.
    file_idx : int
        The index of the scan file used.
    output_folder : str
        The folder where all inSPIRE output is written.
    """
    if file_idx == 0:
        combined_df.write_csv(
            f'{output_folder}/input_all_features.tab',
            separator='\t',
        )
    else:
        with open(f'{output_folder}/input_all_features.tab', mode='ab') as out_file:
            combined_df.write_csv(
                out_file,
                separator='\t',
                has_header=False,
            )

    return f'{output_folder}/input_all_features.tab'

def write_rescoring_features(
        search_df,
        mods_df,
        config,
    ):
    """ Function to write the percolator/mokapot input DataFrame.

    Parameters
    ----------
    search_df : pd.DataFrame
        The results from the original search DataFrame.
    mods_df : pd.DataFrame
        The DataFrame of ptms.
    config : inspire.config.Config
        The Config object.
    """
    feature_df = create_basic_features(search_df, mods_df)

    if config.use_binding_affinity == 'asFeature':
        mhc_pan_df = read_mhcpan_output(f'{config.output_folder}/mhcpan')
        mhc_pan_df = mhc_pan_df.rename({
            'Peptide': PEPTIDE_KEY,
            'Aff(nM)': 'bindingAffinity'
        })
        mhc_pan_df = mhc_pan_df.with_columns(
            pl.col('bindingAffinity').apply(log10).alias('bindingAffinity')
        )
        mhc_pan_df = mhc_pan_df.select([PEPTIDE_KEY, 'bindingAffinity'])
        feature_df = feature_df.join(
            mhc_pan_df,
            on=PEPTIDE_KEY,
            how='inner'
        )

    psm_id_key = PSM_ID_KEY[config.rescore_method]
    if mods_df.empty:
        mod_weights = {}
    else:
        mod_weights = dict(zip(mods_df[PTM_ID_KEY].tolist(), mods_df[PTM_WEIGHT_KEY].tolist()))

    feature_df = feature_df.with_columns(
        pl.struct([PEPTIDE_KEY, PTM_SEQ_KEY]).apply(
            lambda x : modify_sequence_for_skyline(x, mod_weights),
            skip_nulls=False,
        ).alias('modSeq')
    )

    feature_df = feature_df.with_columns(
        pl.struct([SOURCE_KEY, SCAN_KEY, 'modSeq']).apply(
            lambda x : str(x[SOURCE_KEY]) + '_' + str(x[SCAN_KEY]) + '_' + str(x['modSeq']),
            skip_nulls=False,
        ).alias(psm_id_key)
    )
    print(
        OKCYAN_TEXT +
        '\tBasic Features added, adding Spectral Features.' +
        ENDC_TEXT
    )

    write_with_spectral_features(
        feature_df,
        mods_df,
        config,
    )

def process_unknown_modifications(target_df, mods_df, config):
    """ Function to handle modifications which are unknown to the Prosit spectral predictor
        (as well as unmodified cysteines).

    Parameters
    ----------
    Parameters
    ----------
    search_df : pd.DataFrame
        The results from the original search DataFrame.
    mods_df : pd.DataFrame
        The DataFrame of ptms.
    config : inspire.config.Config
        The Config object.

    Returns
    -------
    target_df : pd.DataFrame
        The results from the original search DataFrame with unknown modifications filtered.
    """
    if config.drop_unknown_mods:
        unknown_mods = mods_df[
            (mods_df[PTM_NAME_KEY] != 'Oxidation (M)') &
            (mods_df[PTM_NAME_KEY] != 'Carbamidomethylation') &
            (mods_df[PTM_NAME_KEY] != 'Carbamidomethyl (C)')
        ][PTM_ID_KEY].tolist()
        unknown_mods = {str(x) for x in unknown_mods}
        target_df = target_df.with_columns(
            pl.col(PTM_SEQ_KEY).apply(
                lambda x : check_bad_mods(x, unknown_mods),
                skip_nulls=False,
            ).alias('unknownModifications')
        )
        count_before_drop = target_df.shape[0]
        if config.use_accession_stratum:
            target_df = accession_informed_filter(target_df, 'unknownModifications')
        else:
            target_df = target_df.filter(
                pl.col('unknownModifications').is_not()
            ).drop('unknownModifications')

        count_after_drop = target_df.shape[0]
        filtered_psms = count_before_drop - count_after_drop
        print(
            OKCYAN_TEXT +
            f'\tFiltered {filtered_psms} PSMs due to modifications unknown to Prosit.' +
            ENDC_TEXT
        )

        if not mods_df[
                (mods_df[PTM_NAME_KEY] == 'Carbamidomethylation') |
                (mods_df[PTM_NAME_KEY] == 'Carbamidomethyl (C)')
            ].shape[0] and config.filter_c:
            count_before_drop = target_df.shape[0]
            target_df = target_df.filter(
                ~pl.col(PEPTIDE_KEY).str.contains('C')
            )
            count_after_drop = target_df.shape[0]
            filtered_psms = count_before_drop - count_after_drop
            print(
                OKCYAN_TEXT +
                f'\tFiltered {filtered_psms} PSMs due to unmodified cysteines.' +
                ENDC_TEXT
            )

    return target_df

def create_features(config):
    """ Function to create features for percolator/mokapot input.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object used throughout the pipeline.
    """
    target_df, mods_df = generic_read_df(config)
    target_df = target_df.with_row_count(name='tempIndex')

    target_df = process_unknown_modifications(target_df, mods_df, config)

    n_variable_mods = mods_df.shape[0]  # pylint: disable=no-member
    if  n_variable_mods > 9:
        raise ValueError(
            f'inSPIRE supports no more than 9 unique PTMs, found {n_variable_mods}'
        )

    print(
        OKCYAN_TEXT +
        '\tMS Search Results ready.' +
        ENDC_TEXT
    )
    write_rescoring_features(
        target_df,
        mods_df,
        config,
    )



""" Functions for generating the html report at the end of the PEPSeek
    pipeline.
"""
import os
import webbrowser

import pandas as pd

from inspire.constants import ENDC_TEXT, OKCYAN_TEXT

def create_host_report(config):
    """ Function to create the final html report and open it in the brower.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the whole pipeline.
    figures : dict
        A dictionary containing all of the required plots.
    """
    out_path = os.path.abspath(config.output_folder)

    base_table = """<table style="width:40%">
        <tr>
            <th>Peptide</th>
            <th>Proteins</th>
        </tr>
    """
    with open(f'{out_path}/img/PEPSeek_host_bar_plot.svg', mode='r', encoding='UTF-8') as in_f:
        bar_plot = in_f.read().strip('\n')

    with open(f'{out_path}/img/PEPSeek_host_metrics.svg', mode='r', encoding='UTF-8') as in_f:
        swarm_plot = in_f.read()

    base_table += '</table>'

    html_string = ('''
    <html>
        <head>
            <link 
                rel="stylesheet"
                href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.1/css/bootstrap.min.css"
            >
            <style>
                body{
                    font-family: Helvetica;
                    margin:0 100;
                    padding-bottom: 50px;
                    background:whitesmoke;
                }
                h2{
                    color: firebrick;
                    font-family: Helvetica;
                }
                h3{
                    color: firebrick;
                    font-family: Helvetica;
                }
                h2:after
                {
                    content:' ';
                    display: block;
                    border:0.25px solid #696969;
                    position: absolute;
                    width: 60%;
                    margin-top: 2px;
                    left: 20%;
                }
                table {
                    font-family: Helvetica;
                    width: 60%;
                    border: 2px solid #696969;
                }
                th, td {
                    border: 1px solid #696969;
                    padding: 2px;
                }
            </style>
        </head>
        <body>
            <center>
            <h2>PEPSeek Host Peptide Report for ''' + config.experiment_title + '''</h2>
            </center>
            <p>
                Note quantitative information on host peptide changes upon infection can be
                found in the quantitative report.
            </p>
            <h3>
                Host Peptide Candidates Found
            </h3>
            <p>
                This figure compares the number of peptides identified by both the
                search engine and PEPSeek against peptides identified by PEPSeek alone.
            </p>
            <br>
             ''' + bar_plot +
    '''
            <h3>
                MS2 Quality Metrics
            </h3>
            <p>
                This figure shows distribution of spectral angle, engine score, and
                retention time prediction error on the host peptide identifications.
                For spectral angle and retention time prediction error we should see a
                similar distribution. On engine score, the PEPSeek only peptides likely
                score lower.
            </p>
            <br>
    ''' + swarm_plot
    )
    if os.path.exists(f'{out_path}/img/logo_comp_plots.svg'):
        with open(f'{out_path}/img/logo_comp_plots.svg', mode='r', encoding='UTF-8') as in_f:
            logo_plot = in_f.read().strip('\n')
        html_string += (
            '''
                    <h3>
                        Pathogen vs. Host Amino Acid Frequency:
                    </h3>
                    <p>
                        This chart shows the JS divergence between amino acid frequencies in host
                        and pathogen peptides. Amino acids on the positive y-axis are
                        overrepresented in the pathogen peptides compared to the host,
                        while amino acids on the negative y-axis
                        are overrepresented in the host peptides compared to the pathogen.
                        This plot can provide insight into the differing characteristics
                        of pathogen and host peptides.
                    </p>
            ''' + logo_plot
        )

    if os.path.exists(f'{out_path}/img/PEPSeek_affinity_cluster.svg'):
        with open(
            f'{out_path}/img/PEPSeek_host_affinity_cluster.svg',
            mode='r',
            encoding='UTF-8',
        ) as in_f:
            aff_clust = in_f.read().strip('\n')
        html_string += (
            '''
                <h3>
                    Binding Affinity Predictions:
                </h3>
                <p>
                    The predicted binding affinities for the identified host peptides. This can
                    help see you the distribution of predicted binders for each allele in
                    your host cell. Colour is based on NetMHCpan predicted binding affinty
                    (percentage rank).
                </p>
            ''' + aff_clust
        )

    output_path = f'{config.output_folder}/PEPSeek/pepseek-host-report.html'
    with open(output_path, 'w', encoding='UTF-8') as output_file:
        output_file.write(html_string)

    print(
        OKCYAN_TEXT +
        '\tPEPSeek host report generated.' +
        ENDC_TEXT
    )

    if not config.silent_execution:
        webbrowser.open(
            'file://' + os.path.realpath(output_path),
            new=2
        )



""" Helpful functions used across the module.
"""
import multiprocessing as mp
import pickle
import re

from Bio import SeqIO
import pandas as pd
import polars as pl

from inspire.constants import (
    ACCESSION_STRATUM_KEY,
    CHARGE_KEY,
    ENGINE_SCORE_KEY,
    KNOWN_PTM_LOC,
    KNOWN_PTM_WEIGHTS,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_IS_VAR_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    PTM_WEIGHT_KEY,
    SCAN_KEY,
    SOURCE_KEY,
    SPECTRAL_ANGLE_KEY,
)
from inspire.input.mgf import process_mgf_file
from inspire.input.mzml import process_mzml_file

def fetch_proteome(proteome, with_desc=True):
    """ Function to read in proteome fasta file and return list of tuples containing:
        0. protein name
        1. protein sequence
        2. protein description (optional)
    """
    if with_desc:
        return [
            (x.name, str(x.seq).replace('I', 'L'), x.description) for x in SeqIO.parse(
                proteome, 'fasta',
            )
        ]
    return [
        (x.name, str(x.seq).replace('I', 'L')) for x in SeqIO.parse(
            proteome, 'fasta',
        )
    ]

def parallel_remap(combined_df, n_cores, proteome, out_column, trace_accession=True):
    """ Function to remap peptides in a DataFrame to a proteome in parallel.

    Parameters
    ----------
    combined_df : pl.DataFrame
        DataFrame including peptides identified.
    n_cores : int
        The number of CPUs that should be used in parallel.
    proteome : list of tuple
        The liste of proteins (position 0 is name and position 1 is sequence).
    out_column : str
        The name of the column
    trace_accession : bool (default=True)
        Flag indicating if full accession should be trace or just a boolean flag if
        peptide is found in proteome.

    Returns
    -------
    combined_df : pl.DataFrame
        DataFrame including peptides identified with remapped accessions.
    """
    pep_df = combined_df.select(PEPTIDE_KEY).unique()
    batch_size = pep_df.shape[0]//n_cores
    batch_values = []
    for batch_idx in range(n_cores):
        batch_values.extend([batch_idx]*batch_size)

    if (additional_psms := pep_df.shape[0]%n_cores):
        batch_values.extend([n_cores-1]*additional_psms)

    pep_df = pep_df.with_columns(
        pl.Series(name='batch', values=batch_values)
    )
    pep_df_list = pep_df.partition_by('batch')
    func_args = []
    for sub_pep_df in pep_df_list:
        func_args.append([sub_pep_df, proteome, out_column, trace_accession])

    with mp.get_context('spawn').Pool(processes=n_cores) as pool:
        pep_df_list = pool.starmap(_sub_remap, func_args)
    remapped_pep_df = pl.concat(pep_df_list)

    combined_df = combined_df.join(remapped_pep_df, how='inner', on='peptide')

    return combined_df

def _sub_remap(pep_df, proteome, out_column, trace_accession):
    pep_df = pep_df.with_columns(
        pl.col('peptide').apply(
            lambda x : remap_to_proteome(x, proteome, trace_accession=trace_accession)
        ).alias(out_column)
    )
    return pep_df


def remap_to_proteome(
        peptide,
        proteome,
        trace_accession=True
    ):
    """ Function to check for the presence of an identified peptide as either canonical
        or spliced in the input proteome.
    """
    il_peptide = peptide.replace('I', 'L')
    accession_stratum = 'unknown'
    for protein in proteome:
        if il_peptide in protein[1]:
            if not trace_accession:
                return True
            if accession_stratum == 'unknown':
                accession_stratum = protein[0]
            else:
                accession_stratum += f' {protein[0]}'

    if not trace_accession:
        return False
    return accession_stratum

def get_mod_score(df_row):
    """ Helper function to compare different accession strata.
    """
    if df_row[ACCESSION_STRATUM_KEY] == 0:
        return df_row[ENGINE_SCORE_KEY]
    return 0.7*df_row[ENGINE_SCORE_KEY]

def accession_informed_filter(target_df, drop_feature_key):
    """ Function for filtering with competitors from lower accession strata excluded.
    """
    target_df = target_df.with_columns(
        pl.struct(ENGINE_SCORE_KEY, ACCESSION_STRATUM_KEY).apply(
            get_mod_score
        ).alias('modEngineScore')
    )

    target_df = target_df.with_columns(
        pl.col('modEngineScore').max().over(
            [SOURCE_KEY, SCAN_KEY]
        ).alias('maxModScore')
    )

    top_rank_mod_df = target_df.filter(
        (pl.col(drop_feature_key)) &
        (pl.col('maxModScore').eq(pl.col('modEngineScore')))
    )
    top_rank_mod_df = top_rank_mod_df.select([SOURCE_KEY, SCAN_KEY]).unique()
    top_rank_mod_df = top_rank_mod_df.with_columns(
        pl.lit('yes').alias('drop')
    )
    target_df = target_df.filter(
        pl.col(drop_feature_key).is_not()
    ).drop(['modEngineScore', 'maxModScore', drop_feature_key])

    target_df = target_df.join(
        top_rank_mod_df,
        how='left',
        on=[SOURCE_KEY, SCAN_KEY]
    )
    target_df = target_df.filter(pl.col('drop').ne('yes'))
    target_df = target_df.drop('drop')

    return target_df

def fetch_collision_energy(output_folder):
    """ Function to fetch the calibrated collision energy setting.
    """
    results_df = pd.read_csv(
        f'{output_folder}/collisionEnergyStats.csv'
    )
    optimal_collision_energy = results_df['collisionEnergy'].iloc[
        results_df[SPECTRAL_ANGLE_KEY].idxmax()
    ]

    return optimal_collision_energy

def permute_ptms(peptide, ptm_seq, uniform_length=False):
    """ Function to generate all possible permutations on the PTMs of a peptide
        sequence due to adjacent amino acid swaps.

    Parameters
    ----------
    peptide : str
        The original peptide.
    ptm_seq : str or nan
        The PTM sequence of the peptide.
    uniform_length : bool (default=False)
        Flag indicating whether to return a list of length 29 (max possible permutations)
        or to only return the permutations possible.

    Returns
    -------
    permed_ptms : list of str or None
        A list of the possible PTM permutations.
    """
    if not isinstance(ptm_seq, str):
        return [ptm_seq]*29

    permed_ptms = []
    for idx in range(len(ptm_seq)-5):
        if peptide[idx] != peptide[idx+1]:
            permed_ptms.append(
                ptm_seq[:idx+2] + ptm_seq[idx+3] + ptm_seq[idx+2] + ptm_seq[idx+4:]
            )
        elif uniform_length:
            permed_ptms.append(None)
    if uniform_length:
        permed_ptms += [None]*(29-len(permed_ptms))
    return permed_ptms

def permute_seq(peptide, uniform_length=False):
    """ Function to generate all possible permutations on a peptide
        sequence due to adjacent amino acid swaps.

    Parameters
    ----------
    peptide : str
        The original peptide.
    uniform_length : bool (default=False)
        Flag indicating whether to return a list of length 29 (max possible permutations)
        or to only return the permutations possible.

    Returns
    -------
    permed_peps : list of str or None
        A list of the possible peptide permutations.
    """
    permed_peps = []
    for idx in range(len(peptide)-1):
        if peptide[idx] != peptide[idx+1]:
            permed_peps.append(
                peptide[:idx] + peptide[idx+1] + peptide[idx] + peptide[idx+2:]
            )
        elif uniform_length:
            permed_peps.append(None)
    if uniform_length:
        permed_peps += [None]*(29-len(permed_peps))
    return permed_peps

def check_bad_mods(ptm_str, bad_mods):
    """ Function to check for the presence of ptms other than oxidation of methionine
        or carbamidomehtylation of Cysteine.
    """
    if not isinstance(ptm_str, str):
        return False

    for mod_id in ptm_str:
        if mod_id in bad_mods:
            return True
    return False

def get_cam_flag(mods_df):
    """ Function to get the flag for oxidation of methionine from the PTMs DataFrame.
    Parameters
    ----------
    mods_df : pd.DataFrame
        A small DataFrame detailing the PTMs found in the data.
    Returns
    -------
    carb_flag : int
        The flag for carbamidomtheylation of cysteine.
    """
    try:
        carb_flag = int(mods_df[
            (mods_df[PTM_NAME_KEY] == 'Carbamidomethylation') |
            (mods_df[PTM_NAME_KEY] == 'Carbamidomethyl (C)')
        ][PTM_ID_KEY].iloc[0])
    except IndexError:
        carb_flag = -2

    return carb_flag

def get_ox_flag(mods_df):
    """ Function to get the flag for oxidation of methionine from the PTMs DataFrame.

    Parameters
    ----------
    mods_df : pd.DataFrame
        A small DataFrame detailing the PTMs found in the data.

    Returns
    -------
    ox_flag : int
        The flag for oxidation of methionine.
    """
    try:
        ox_flag = int(mods_df[mods_df[PTM_NAME_KEY] == 'Oxidation (M)'][PTM_ID_KEY].iloc[0])
    except IndexError:
        ox_flag = -1

    return ox_flag

def get_mokapot_weights(output_folder, inspire_step):
    """ Function to get the weights of mokapot models.

    Parameters
    ----------
    output_folder : str
        The folder where all inSPIRE output is written.
    inspire_step : str
        Indicator of which step fo inSPIRE the model was trained for.

    Results
    -------
    weights_df : pd.DataFrame
        A DataFrame of the weights for different features.
    """
    for fold_idx in range(1, 4):
        with open(
            f'{output_folder}/{inspire_step}.mokapot.model_fold-{fold_idx}.pkl',
            'rb',
        ) as model_file:
            model = pickle.load(model_file)
        weights = model.estimator.coef_[0]
        if fold_idx == 1:
            n_feats = len(weights)
            feat_weights = {f'feat{idx}': [] for idx in range(1, n_feats+1)}
        for feat_idx, name in enumerate(feat_weights):
            feat_weights[name].append(round(weights[feat_idx], 2))

    return pd.DataFrame(feat_weights)

def modify_sequence_for_skyline(df_row, mod_weights):
    """ Helper function to modify the peptide sequence so that it is compatitable
        with Skyline input.

    Parameters
    ----------
    df_row : pd.Series
        A row of the DataFrame.
    mod_weights : dict
        A dictionary mapping ptm IDs to molecular weights.

    Returns
    -------
    modified_sequence : str
        The modified sequence in the correct format for Skyline.
    """
    if not isinstance(df_row[PTM_SEQ_KEY], str) or not df_row[PTM_SEQ_KEY]:
        return df_row[PEPTIDE_KEY]

    modified_sequence = ''
    for idx, entry in enumerate(df_row[PEPTIDE_KEY]):
        mod = df_row[PTM_SEQ_KEY][idx+2]
        if mod != '0':
            mod_wt = mod_weights[int(mod)]
            if mod_wt > 0:
                modified_sequence += f'{entry}[+{round(mod_wt, 1)}]'
            else:
                modified_sequence += f'{entry}[{round(mod_wt, 1)}]'
        else:
            modified_sequence += entry
    return modified_sequence

def read_distiller_log(distiller_log):
    """ Function to read the source file names from a distiller log.

    Parameters
    ----------
    distiller_log : str
        The location of distiller output giving original source names.
    """
    source_files = []
    with open(distiller_log, 'r', encoding='UTF-8') as distiller_output:
        while (line := distiller_output.readline()):
            if line.startswith('Raw file '):
                source = line.split('/')[-1].split('\\')[-1].split('.raw')[0]
                source_files.append(source)

    return source_files

def remove_source_suffixes(source):
    """ Helper function to remove raw, mzML or mgf suffixes from source name.

    Parameters
    ----------
    source : str
        The name of a source file.

    Returns
    -------
    source : str
        The updated name with a suffix removed.
    """
    if source.endswith('.mzML'):
        return source[:-5]
    if source.endswith('.raw') or source.endswith('.mgf'):
        return source[:-4]
    return source


def filter_for_prosit(search_df):
    """ Function to filter sequences not suitable for Prosit input (polars DataFrame).

    Parameters
    ----------
    search_df : pl.DataFrame
        A DataFrame of search results from an ms search engine.

    Returns
    -------
    search_df : pl.DataFrame
        The input DataFrame with sequences not suitable for Prosit input removed.
    """
    search_df = search_df.filter(
        pl.col(PEPTIDE_KEY).is_not_null() & (pl.col(CHARGE_KEY).is_not_null())
    )
    search_df = search_df.filter(
        pl.col(PEPTIDE_KEY).apply(
            lambda x : isinstance(x, str) and 'U' not in x and len(x) > 6 and len(x) < 31,
            skip_nulls=False,
        )
    )
    search_df = search_df.filter(
        pl.col(CHARGE_KEY).lt(7)
    )

    return search_df

def _modify_ptm_seq(pep_seq, ptm_seq, ptm_locs, ptm_id):
    """ Helper function to add a fixed PTM to the ptm_seq column of MS search
        result DataFrame.

    Parameters
    ----------
    pep_seq : str
        The peptide sequence.
    ptm_seq : str
        The ptm sequence.
    ptm_locs : str
        The location(s) where the ptm should be added.
    ptm_id : str
        The id of the ptm.

    Returns
    -------
    ptm_seq : str
        The modified ptm sequence.
    """
    if ptm_locs is None:
        return ptm_seq

    if ptm_locs == 'N-term':
        if not isinstance(ptm_seq, str):
            return ptm_id + '.' + ''.join(['0']*len(pep_seq)) + '.0'
        return ptm_id + ptm_seq[1:]


    for loc in ptm_locs:
        edit_points = [m.start() for m in re.finditer(loc, pep_seq)]
        if edit_points and not isinstance(ptm_seq, str):
            ptm_seq = '0.' + ''.join(['0']*len(pep_seq)) + '.0'
        for e_point in edit_points:
            ptm_seq = ptm_seq[:e_point+2] + ptm_id + ptm_seq[e_point+3:]

    return ptm_seq

def add_fixed_modifications(search_df, mods_df, fixed_modifications):
    """ Function to add fixed modifications to the search data.

    Parameters
    ----------
    search_df : pd.DataFrame
        The DataFrame of MS search results from either MaxQuant or PEAKS.
    mods_df : pd.DataFrame
        The DataFrame of variable modifications in the search results.
    fixed_modifications : list
        A list of the fixed modifications found in the DataFrame.

    Returns
    -------
    mods_df : pd.DataFrame
        A DataFrame of all modifications found (both variable and fixed).
    """
    ptm_ids = []
    ptm_names = []
    ptm_weights = []
    for idx, modification in enumerate(fixed_modifications):
        if modification in mods_df[PTM_NAME_KEY]:
            continue

        ptm_weight = KNOWN_PTM_WEIGHTS.get(modification)
        ptm_idx = mods_df.shape[0]+idx+1
        if ptm_weight is None:
            raise ValueError(f'Unsupported fixed modification {modification}.')

        ptm_ids.append(ptm_idx)
        ptm_names.append(modification)
        ptm_weights.append(ptm_weight)
        str_ptm_idx = str(ptm_idx)
        search_df = search_df.with_columns(
            pl.struct([PEPTIDE_KEY, PTM_SEQ_KEY]).apply(
                lambda x, mod=modification, ptm_id=str_ptm_idx : _modify_ptm_seq(
                    x[PEPTIDE_KEY],
                    x[PTM_SEQ_KEY],
                    KNOWN_PTM_LOC.get(mod),
                    ptm_id
                )
            ).alias(PTM_SEQ_KEY)
        )

    fixed_ptm_df = pd.DataFrame({
        PTM_ID_KEY: ptm_ids,
        PTM_NAME_KEY: ptm_names,
        PTM_WEIGHT_KEY: ptm_weights,
        PTM_IS_VAR_KEY: [False]*len(fixed_modifications)
    })
    mods_df = pd.concat([
        mods_df[[PTM_ID_KEY, PTM_NAME_KEY, PTM_WEIGHT_KEY, PTM_IS_VAR_KEY]],
        fixed_ptm_df]
    )

    return search_df, mods_df

def convert_mod_seq_to_ptm_seq(mod_seq):
    """ Function to convert the modified prosit sequence to the PTM seq used elsewhere.

    Parameters
    ----------
    mod_seq : str
        The input sequence for Prosit.

    Returns
    -------
    ptm_seq : str
        The string of digits used to represent any PTMs present.
    """
    ptm_seq = '0.'
    while mod_seq:
        if len(mod_seq) == 1 or mod_seq[1] != '[':
            ptm_seq += '0'
            mod_seq = mod_seq[1:]
        elif mod_seq[1] == '[' and mod_seq[0] == 'C':
            ptm_seq += '2'
            mod_seq = mod_seq[8:]
        else:
            ptm_seq += '1'
            mod_seq = mod_seq[8:]

    ptm_seq += '.0'
    return ptm_seq

def fetch_scan_data(input_df, config, with_charge, with_rt=False):
    """ Function to fetch the experimental scan data.

    Parameters
    ----------
    input_df : pd.DataFrame or pl.DataFrame
        The DataFrame of PSMs whose spectra we wish to plot.
    config : inspire.config.Config
        The Config object used to run the inSPIRE experiment.

    Returns
    -------
    total_scan_df : pl.DataFrame
        A DataFrame of the necessary scan data.
    """
    if isinstance(input_df, pd.DataFrame):
        source_files = input_df[SOURCE_KEY].unique().tolist()
    else:
        source_files = input_df[SOURCE_KEY].unique().to_list()

    scan_dfs = []
    if config.combined_scans_file is not None:
        scan_ids = input_df[SCAN_KEY].tolist()
        mgf_filename = f'{config.scans_folder}/{config.combined_scans_file}'
        scan_dfs = [process_mgf_file(
            mgf_filename,
            set(scan_ids),
            config.scan_title_format,
            config.source_files,
            combined_source_file=True,
            with_charge=with_charge,
            with_retention_time=with_rt,
        )]
    else:
        for scan_file in source_files:
            if isinstance(input_df, pd.DataFrame):
                scan_ids = input_df[input_df[SOURCE_KEY] == scan_file][SCAN_KEY].tolist()
            else:
                scan_ids = input_df.filter(input_df[SOURCE_KEY] == scan_file)[SCAN_KEY].to_list()

            if config.scans_format == 'mzML':
                scan_df = process_mzml_file(
                    f'{config.scans_folder}/{scan_file}.{config.scans_format}',
                    set(scan_ids),
                    with_charge=with_charge,
                    with_retention_time=with_rt,
                )
            else:
                mgf_filename = f'{config.scans_folder}/{scan_file}.{config.scans_format}'
                scan_df = process_mgf_file(
                    mgf_filename,
                    set(scan_ids),
                    config.scan_title_format,
                    config.source_files,
                    combined_source_file=False,
                    with_charge=with_charge,
                    with_retention_time=with_rt,
                )
            scan_dfs.append(scan_df)

    total_scan_df = pl.concat(scan_dfs)

    return total_scan_df

def is_control(source, control_flags):
    """ Function to check if control flags are found in the source files where
        a peptide was identified.

    sources : list of str
        A list of source files from which a single peptide was identified.
    control_flags : list of str
        A list of the control flags which mark a source file as a control file.
    """
    for c_flag in control_flags:
        if c_flag in source:
            return True

    return False



""" Function for investigating IP samples containing potential spliced epitopes.
"""
from copy import deepcopy
from math import floor
import shutil
import os

import pandas as pd

from inspire.constants import (
    ENGINE_SCORE_KEY,
    FINAL_SCORE_KEY,
    FINAL_POSTEP_KEY,
    OUT_POSTEP_KEY,
    PEPTIDE_KEY,
    SCAN_KEY,
    SOURCE_KEY,
    SPECTRAL_ANGLE_KEY,
)
from inspire.pepseek.proteome_mapping import filter_pathogen_only_peptides
from inspire.pepseek.plot_utils import (
    bar_plot,
    check_engine_results,
    js_divergence,
    plot_binding_clustermap,
    plot_quant_pca,
    swarm_plots,
)
from inspire.pepseek.report_template import create_epitope_report
from inspire.pepseek.host_report_template import create_host_report
from inspire.plot_spectra.plot_spectra import plot_spectra
from inspire.utils import is_control

ANTIGEN_KEY = 'protein'
ANTIGEN_LOC_KEY = 'proteinLocation'
ANTIGEN_DESC_KEY = 'proteinDescription'
EPITOPE_CONF_KEY = 'confidence'
EPITOPE_N_FILES = 'numberOfUniqueFiles'
BASE_REPORTING_COLUMNS = [
    ANTIGEN_KEY,
    ANTIGEN_LOC_KEY,
    ANTIGEN_DESC_KEY,
    FINAL_SCORE_KEY,
    EPITOPE_CONF_KEY,
    ENGINE_SCORE_KEY,
    SPECTRAL_ANGLE_KEY,
    'deltaRT',
]


def extract_epitope_candidates(config):
    """ Function for extracting epitope candidates from inSPIRE identifications.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object which controls the experiment.
    """
    print(config.epitope_cut_level)
    if config.epitope_cut_level == 'psm':
        final_df = pd.read_csv(f'{config.output_folder}/finalPsmAssignments.csv')
    else:
        final_df = pd.read_csv(f'{config.output_folder}/finalPeptideAssignments.csv')

    final_df = final_df[final_df['qValue'] < 0.01]

    # Add columns of interest
    final_df['peptideLength'] = final_df['peptide'].apply(len)
    final_df[EPITOPE_CONF_KEY] = (
        (1 - final_df[FINAL_POSTEP_KEY])*100
    ).apply(
        lambda x : floor(x*100)/100
    )

    # Initial filters:
    final_df = final_df[
        (final_df[FINAL_POSTEP_KEY] < config.epitope_candidate_cut_off) &
        (final_df['peptideLength'] <= config.epitope_length_cut_off)
    ]

    if config.epitope_cut_level == 'peptide':
        final_psm_df = pd.read_csv(f'{config.output_folder}/finalPsmAssignments.csv')
        final_psm_df = final_df[final_df['qValue'] < 0.01]
        final_psm_df = final_df[
            (final_df[FINAL_POSTEP_KEY] < config.epitope_candidate_cut_off) &
            (final_df['peptideLength'] <= config.epitope_length_cut_off)
        ]
        id_cols = ['source', 'scan', 'peptide']
        pep_df_cols = ['postErrProb', 'qValue']
        final_df = pd.merge(
            final_df[id_cols + pep_df_cols], final_psm_df.drop(pep_df_cols, axis=1),
            how='inner', on=id_cols
        )

    # Add accession data:
    final_df, multi_mapped_df, host_df = filter_pathogen_only_peptides(final_df, config)

    # Write peptides which map to both pathogen and host proteomes (not as epitope
    # candidates but potentially interesting)
    write_excluded_peptides(
        multi_mapped_df,
        config,
        'multi_mapped_peptides',
    )

    # Filter those peptides found in control.
    final_df['foundInControl'] = final_df.groupby(PEPTIDE_KEY)[SOURCE_KEY].transform(
        lambda source : found_in_control(source, config.control_flags)
    )
    write_excluded_peptides(
        final_df[final_df['foundInControl']],
        config,
        'control_excluded_peptides',
    )
    final_df = final_df[~final_df['foundInControl']]

    # Write best MS2 spectrum per peptide for plotting later.
    final_df.drop_duplicates(subset=[PEPTIDE_KEY]).to_csv(
        f'{config.output_folder}/plotData.csv', index=False,
    )

    # Add useful columns for number of files where peptide identified and found by search engine.
    final_df[EPITOPE_N_FILES] = final_df.groupby(
        PEPTIDE_KEY
    )[SOURCE_KEY].transform('nunique')
    final_df = check_engine_results(final_df, config)

    # Separate and rename protein ID data
    final_df[ANTIGEN_LOC_KEY] = final_df['pathogenAccessions'].apply(
        lambda x : ','.join([str(y[1]) for y in x])
    )
    final_df[ANTIGEN_DESC_KEY] = final_df['pathogenAccessions'].apply(
        lambda x : ','.join([y[2] for y in x])
    )
    final_df = final_df.rename(columns={'pathogenProteins': ANTIGEN_KEY})

    # Select columns to be reported and drop duplicate PSMs
    final_df = final_df.sort_values(by=FINAL_SCORE_KEY, ascending=False)
    final_peptide_df = final_df[
        [
            PEPTIDE_KEY, EPITOPE_N_FILES, 'foundBySearchEngine'
        ] + BASE_REPORTING_COLUMNS + sorted(
            [
                x for x in final_df.columns if (
                    x.endswith('Affinity') or
                    x.endswith('BindLevel') or
                    x.endswith('%Rank_BA')
                )
            ]
        )
    ].drop_duplicates(subset='peptide')
    final_peptide_df = final_peptide_df.reset_index(drop=True)

    for feature in (ENGINE_SCORE_KEY, SPECTRAL_ANGLE_KEY, 'deltaRT'):
        final_peptide_df[feature] = final_peptide_df[feature].apply(
            lambda x : round(x, 2)
        )

    # Add all data to single csv file and write separated data to csv.
    final_peptide_df.to_csv(
        f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv',
        index=False,
    )
    write_excel_report(final_peptide_df, final_df, config)

    try:
        js_divergence(config, host_df)
    except Exception as e:
        print(f'JS Divergence plotting failed with exception {e}')

    bar_plot(config)
    bar_plot(config, host=True)
    swarm_plots(config)
    swarm_plots(config, host=True)

    try:
        plot_binding_clustermap(config)
        plot_binding_clustermap(config, host=True)
    except Exception as e:
        print(f'Binding cluster map plotting failed with exception {e}')        

    # Provide MS2 spectral plots for identified peptides.
    plot_spectra(config)
    if os.path.exists(f'{config.output_folder}/spectralPlots.pdf'):
        shutil.move(
            f'{config.output_folder}/spectralPlots.pdf',
            f'{config.output_folder}/PEPSeek/spectralPlots.pdf',
        )

    plot_quant_pca(config)

    create_epitope_report(config)
    create_host_report(config)

def write_excluded_peptides(excluded_df, config, file_name):
    """ Function to write peptides which are excluded either due to their presence in
        control files or multi-mapping between host and pathogen.
    """
    if not excluded_df.shape[0]:
        excluded_df.to_csv(
            f'{config.output_folder}/PEPSeek/{file_name}.csv', index=False
        )
        return

    sources = excluded_df[SOURCE_KEY].unique().tolist()
    excluded_df = excluded_df.groupby(
        ['peptide', 'proteins'], as_index=False
    )['source'].apply(list)
    if not excluded_df.shape[0]:
        return

    for source in sources:
        excluded_df[f'{source}_identified'] = excluded_df['source'].apply(
            lambda x, src=source : int(src in x)
        )

    excluded_df['nInfectedFiles'] = excluded_df[[
        f'{source}_identified'for source in sources if not is_control(source, config.control_flags)
    ]].sum(axis=1)

    excluded_df['nControlFiles'] = excluded_df[[
        f'{source}_identified'for source in sources if is_control(source, config.control_flags)
    ]].sum(axis=1)

    excluded_df[
        [PEPTIDE_KEY, 'proteins', 'nInfectedFiles', 'nControlFiles'] +
        [f'{source}_identified' for source in sources]
    ].to_csv(
        f'{config.output_folder}/PEPSeek/{file_name}.csv', index=False
    )

def found_in_control(sources, control_flags):
    """ Function to check if control flags are found in the source files where
        a peptide was identified.

    sources : list of str
        A list of source files from which a single peptide was identified.
    control_flags : list of str
        A list of the control flags which mark a source file as a control file.
    """
    if control_flags is None:
        return False

    for source in sources:
        if is_control(source, control_flags):
            return True
    return False


def write_excel_report(final_peptide_df, final_spectra_df, config):
    """ Function to write a full report of the identified pathogen peptides in Microsoft Excel.

    Parameters
    ----------
    final_df : pd.DataFrame
        The DataFrame of all identified pathogen peptides.
    """
    if final_peptide_df is None:
        # Empty Excel file to be written.
        xl_writer = pd.ExcelWriter( # pylint: disable=abstract-class-instantiated
            f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.xlsx',
            engine='xlsxwriter',
        )
        return

    reporting_columns = [
        PEPTIDE_KEY,
        ANTIGEN_KEY,
        ANTIGEN_LOC_KEY,
        ANTIGEN_DESC_KEY,
        'foundBySearchEngine',
        EPITOPE_N_FILES,
        EPITOPE_CONF_KEY,
        ENGINE_SCORE_KEY,
        SPECTRAL_ANGLE_KEY,
        'deltaRT',
    ]

    final_id_df = final_peptide_df[reporting_columns]
    total_df = deepcopy(final_id_df)

    with pd.ExcelWriter( # pylint: disable=abstract-class-instantiated
        f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.xlsx',
        engine='xlsxwriter',
    ) as xl_writer:
        final_id_df.to_excel(
            xl_writer,
            index=False,
            sheet_name='finalCandidates',
        )
        final_spectra_df[[SOURCE_KEY, SCAN_KEY, PEPTIDE_KEY, 'modifiedSequence']].to_excel(
            xl_writer,
            index=False,
            sheet_name='spectraIdentified',
        )

        if config.use_binding_affinity is not None:
            total_df, xl_writer = add_binding_affinity_data(total_df, final_peptide_df, xl_writer)

        if os.path.exists(f'{config.output_folder}/quant/fold_changes.csv'):
            total_df, xl_writer = add_quantification_data(final_peptide_df, xl_writer, config)

        if config.host_only_results is not None:
            total_df, xl_writer = add_host_only_validation(
                config,
                total_df,
                final_spectra_df,
                xl_writer,
            )

        total_df.to_excel(
            xl_writer,
            index=False,
            sheet_name='allInformation',
        )

        xl_writer = add_conditional_formatting(xl_writer)

def add_quantification_data(final_peptide_df, xl_writer, config):
    """ Function to add quantification data to the output .xlsx file.

    Parameters
    ----------
    final_peptide_df : pd.DataFrame
        The final DataFrame of unique peptides which are potential epitope candidates.
    xl_writer : pd.ExcelWriter
        Writer for output .xlsx file
    config : inspire.config.Config
        Config object which controls the experiment.
    """
    fc_df = pd.read_csv(f'{config.output_folder}/quant/fold_changes.csv')
    final_peptide_df['rank'] = final_peptide_df.index
    fc_df = pd.merge(
        final_peptide_df[['rank', 'peptide']],
        fc_df,
        how='left',
        on='peptide',
    )
    fc_df = fc_df.fillna(0.0)

    fc_df = fc_df.sort_values(by='rank')
    fc_df = fc_df.drop('rank', axis=1)

    fc_df['absChange'] = fc_df[['meanAreaInfected', 'meanAreaControl']].apply(
        lambda df_row : 2**df_row['meanAreaInfected'] - 2**df_row['meanAreaControl'],
        axis=1
    )
    fc_df.to_excel(
        xl_writer,
        index=False,
        sheet_name='quantificationData',
    )
    final_peptide_df = final_peptide_df.drop('rank', axis=1)

    total_df = pd.merge(
        final_peptide_df,
        fc_df.drop('peptide', axis=1),
        left_index=True,
        right_index=True
    )

    return total_df, xl_writer


def add_conditional_formatting(xl_writer):
    """ Function
    """
    workbook  = xl_writer.book
    format1 = workbook.add_format({'bg_color': 'orange'})

    worksheet = xl_writer.sheets['finalCandidates']
    worksheet.conditional_format(
        1, 100, 1, 500,
        {'type': 'formula', 'criteria': ''}
    )
    worksheet.conditional_format(
        'G1:G100',
        {
            'type': 'text',
            'criteria': 'containing',
            'value':    'Yes',
            'format':   format1,
        },
    )
    return xl_writer

def add_binding_affinity_data(total_df, final_peptide_df, xl_writer):
    """ Function to add binding affinity prediction data.

    Parameters
    ----------
    total_df : pd.DataFrame
    """
    binding_affinity_df = final_peptide_df[
        ['peptide'] +
        [
            x for x in final_peptide_df.columns if x.endswith(
                'Level'
            )  or x.endswith(
                'Affinity'
            ) or x.endswith(
                '%Rank_BA'
            )
        ]
    ]
    binding_affinity_df.to_excel(
        xl_writer,
        index=False,
        sheet_name='bindingAffinityData',
    )

    total_df = pd.merge(
        final_peptide_df,
        binding_affinity_df.drop('peptide', axis=1),
        left_index=True,
        right_index=True
    )
    return total_df, xl_writer


def add_host_only_validation(config, total_df, final_df, xl_writer):
    """ Function to add validation from identification from a search without pathogen proteome.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object which controls the experiment.
    """
    comp_df = pd.read_csv(
        config.host_only_results
    )

    original_final_df = pd.read_csv(
        f'{config.output_folder}/finalPsmAssignments.csv'
    )
    original_final_df = original_final_df[
        original_final_df['postErrProb'] < config.epitope_candidate_cut_off
    ]
    original_final_df = original_final_df[['source', 'scan', 'peptide']]
    original_final_df['potentialChimera'] = 'Yes'

    comp_df = pd.merge(
        comp_df,
        original_final_df,
        how='left',
        on=['source', 'scan', 'peptide']
    )

    comp_df['potentialChimera'] = comp_df['potentialChimera'].fillna('No')

    comp_df = comp_df[[
        'source', 'scan', 'potentialChimera', 'peptide', 'spectralAngle', 'engineScore',
    ]].rename(
        columns={
            'peptide':  'competitorPeptide',
            'spectralAngle': 'competitorSpectralAngle',
            'engineScore': 'competitorEngineScore',
        }
    )

    final_comp_df = pd.merge(
        final_df[
            ['source', 'scan', 'peptide', FINAL_SCORE_KEY]
        ].drop_duplicates(subset='peptide'),
        comp_df,
        how='left',
        on=['source', 'scan']
    )

    final_comp_df = final_comp_df.sort_values(by=FINAL_SCORE_KEY, ascending=False)

    final_comp_df = final_comp_df.drop(['source', 'scan', FINAL_SCORE_KEY], axis=1)

    final_comp_df.to_excel(
        xl_writer,
        index=False,
        sheet_name='competitorData',
    )

    total_df = pd.merge(
        total_df,
        final_comp_df.drop('peptide', axis=1),
        left_index=True,
        right_index=True
    )

    return total_df, xl_writer



""" Definition of Config class.
"""
import glob
import os
from pathlib import Path
import platform

import yaml

from inspire.utils import read_distiller_log

ALL_CONFIG_KEYS = [
    'accessionFlags',
    'accessionFormat',
    'accessionKeys',
    'alleles',
    'collisionEnergy',
    'fraggerUseContams',
    'combinedScansFile',
    'controlFlags',
    'deltaMethod',
    'distillerLog',
    'dropUnknownPTMs',
    'epitopeCandidateCutOff',
    'epitopeCutLevel',
    'excludeFeatures',
    'engineScoreCut',
    'enzyme',
    'experimentTitle',
    'falseDiscoveryRate',
    'fixedModifications',
    'filterCysteine',
    'forceReload',
    'fraggerMemory',
    'fraggerDbSplits',
    'fraggerPath',
    'fraggerParams',
    'hostOnlyResults',
    'hostProteome',
    'includeFeatures',
    'inferProteins',
    'mapContaminants',
    'ms2pipModel',
    'ms1Accuracy',
    'mzAccuracy',
    'mzRange',
    'mzUnits',
    'nCores',
    'netMHCpan',
    'outputFolder',
    'pathogenProteome',
    'plotSpectraSourceSplit',
    'proteinDecoyKey',
    'proteome',
    'quantCutOff',
    'rtFitLoc',
    'reduce',
    'remapToProteome',
    'replaceIL',
    'rescoreCommand',
    'rescoreMethod',
    'resultsExport',
    'reuseInput',
    'panDocker',
    'scansFolder',
    'scansFormat',
    'scanTitleFormat',
    'searchResults',
    'searchEngine',
    'sourceFileName',
    'silentExecution',
    'skylineConfig',
    'skylineReportTemplate',
    'skylineRunner',
    'skylineIdpCutOff',
    'skylinePrecursorCountFilter',
    'skylineRatioCutOff',
    'spectralAngleDfs',
    'spectralPredictor',
    'useAccessionStrata',
    'useBindingAffinity',
    'useIrtDelta',
    'useMinimalFeatures',
    'technicalReplicates',
]

class Config:
    """ Holder for configuration of the inspire pipeline.
    """
    def __init__(self, config_file):
        with open(config_file, 'r', encoding='UTF-8') as stream:
            config_dict = yaml.safe_load(stream)
        for config_key in config_dict:
            if config_key not in ALL_CONFIG_KEYS:
                raise ValueError(f'Unrecognised key {config_key} found in config file.')
        self._load_data(config_dict)
        self._clean_file_paths()

    def _clean_file_paths(self):
        """ Function to clean the file paths given to inspire.
        """
        home = str(Path.home())

        if isinstance(self.search_results, str):
            self.search_results = self.search_results.replace('~', home).replace(
                '%USERPROFILE%', home
            )
        elif isinstance(self.search_results, list):
            self.search_results = [
                x.replace('~', home).replace('%USERPROFILE%', home) for x in self.search_results
            ]

        self.scans_folder = self.scans_folder.replace('~', home).replace('%USERPROFILE%', home)
        if self.scans_folder.endswith('/'):
            self.scans_folder = self.scans_folder[:-1]
        self.output_folder = self.output_folder.replace('~', home).replace('%USERPROFILE%', home)
        if self.output_folder.endswith('/'):
            self.output_folder = self.output_folder[:-1]

        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder)

        if not os.path.exists(f'{self.output_folder}/quant'):
            os.makedirs(f'{self.output_folder}/quant')

        if not os.path.exists(f'{self.output_folder}/PEPSeek'):
            os.makedirs(f'{self.output_folder}/PEPSeek')

        if not os.path.exists(f'{self.output_folder}/img'):
            os.makedirs(f'{self.output_folder}/img')

    def _load_data(self, config_dict):
        """ Function to load data.
        """
        home = str(Path.home())

        # Required:
        self.experiment_title = config_dict['experimentTitle']
        self.output_folder = config_dict['outputFolder']
        search_res = config_dict.get('searchResults')
        if isinstance(search_res, list):
            self.search_results = []
            for result_group in search_res:
                self.search_results.extend(glob.glob(result_group))
        else:
            self.search_results = search_res
        if (
            self.search_results is None and
            os.path.exists(f'{self.output_folder}/fragger_searches.txt')
        ):
            with open(
            f'{self.output_folder}/fragger_searches.txt',
            mode='r',
            encoding='UTF-8',
        ) as search_file:
                self.search_results = [line.rstrip() for line in search_file]
        self.search_engine = config_dict['searchEngine']
        self.scans_folder = config_dict['scansFolder']
        self.scans_format = config_dict['scansFormat']

        # Recommended:
        self.map_contaminants = config_dict.get('mapContaminants', 'standard')
        self.results_export = config_dict.get('resultsExport', 'psm')
        self.use_irt_diff = config_dict.get('useIrtDelta', True)
        self.proteome = config_dict.get('proteome')
        if self.proteome is None:
            self.infer_proteins = config_dict.get('inferProteins', False)
        else:
            self.infer_proteins = config_dict.get('inferProteins', True)
        self.n_cores = config_dict.get('nCores', 1)
        self.collision_energy = config_dict.get('collisionEnergy', None)
        self.ms1_accuracy = config_dict.get('ms1Accuracy', 5)
        self.mz_accuracy = config_dict.get('mzAccuracy', 0.02)
        self.mz_units = config_dict.get('mzUnits', 'Da')
        self.enzyme = config_dict.get('enzyme', None)
        self.replace_il = config_dict.get('replaceIL', False)
        self.fixed_modifications = config_dict.get('fixedModifications', None)
        self.spectral_predictor = config_dict.get('spectralPredictor', 'prosit')
        if self.spectral_predictor == 'prosit':
            self.delta_method = config_dict.get('deltaMethod', 'predictor')
        else:
            self.delta_method = config_dict.get('deltaMethod', 'ignore')
        self.reuse_input = config_dict.get('reuseInput', False)
        self.minimal_features = config_dict.get('useMinimalFeatures', False)

        # MSFragger
        self.fragger_use_contams = config_dict.get('fraggerUseContams', True)
        self.fragger_memory = config_dict.get('fraggerMemory', 60)
        self.fragger_db_splits = config_dict.get('fraggerDbSplits', 4)
        self.fragger_path = config_dict.get('fraggerPath')
        self.fragger_params = config_dict.get(
            'fraggerParams',
            f'{home}/inSPIRE_models/utilities/fragger_template.params',
        )

        # Quantification
        self.quantification_cut_off = config_dict.get('quantCutOff', 0.01)
        self.technical_replicates = config_dict.get('technicalReplicates')
        self.skyline_runner = config_dict.get('skylineRunner')
        self.skyline_config_file = config_dict.get(
            'skylineConfig', f'{home}/inSPIRE_models/utilities/skyline_config.sky'
        )
        self.skyline_precursor_filter = config_dict.get('skylinePrecursorCountFilter', 3)
        self.skyline_report_template = config_dict.get(
            'skylineReportTemplate', f'{home}/inSPIRE_models/utilities/skyline_report_template.skyr'
        )
        self.skyline_idp_cut_off = config_dict.get('skylineIdpCutOff', 0.5)
        self.skyline_bg_ratio_cut_off = config_dict.get('skylineRatioCutOff', 0.8)

        # Protein Inference
        self.decoy_protein_flag = config_dict.get('proteinDecoyKey', 'rev_')

        # Optional
        self.remap_to_proteome = config_dict.get('remapToProteome', False)
        self.fdr = config_dict.get('falseDiscoveryRate', 0.01)
        self.exclude_features = config_dict.get('excludeFeatures', [])
        self.include_features = config_dict.get('includeFeatures', None)
        self.reduce = config_dict.get('reduce', False)
        self.rescore_method = config_dict.get('rescoreMethod', 'percolator')
        self.rescore_command = config_dict.get('rescoreCommand')
        if self.rescore_command is None:
            if self.rescore_method.startswith('percolator'):
                self.rescore_command = 'percolator'
            else:
                self.rescore_command = 'mokapot'

        self.sa_query_dfs = config_dict.get('spectralAngleDfs', None)
        self.silent_execution = config_dict.get('silentExecution', False)
        self.plot_spectra_source_split = config_dict.get('plotSpectraSourceSplit')

        self.rt_fit_loc = config_dict.get('rtFitLoc', None)

        self.filter_c = config_dict.get('filterCysteine', True)
        if self.spectral_predictor == 'prosit':
            self.drop_unknown_mods = config_dict.get('dropUnknownPTMs', True)
        else:
            self.drop_unknown_mods = config_dict.get('dropUnknownPTMs', False)

        self.force_reload=config_dict.get('forceReload', False)

        # Mascot files
        self.combined_scans_file = config_dict.get('combinedScansFile', None)
        self.scan_title_format = config_dict.get('scanTitleFormat', None)
        self.source_files = None
        if 'distillerLog' in config_dict:
            self.source_files = read_distiller_log(config_dict['distillerLog'])
        self.source_filename = config_dict.get('sourceFileName', None)

        # NetMhcPan
        self.pan_docker = config_dict.get(
            'panDocker',
            platform.system() == 'Windows',
        )
        self.use_binding_affinity = config_dict.get('useBindingAffinity', None)
        self.pan_command = config_dict.get('netMHCpan')
        default_ba_pred_limit = 15
        if self.use_binding_affinity == 'asFeature':
            default_ba_pred_limit = 31
        self.ba_pred_limit = config_dict.get('baPredictionLimit', default_ba_pred_limit)

        # MS2PIP Model
        self.ms2pip_model = config_dict.get('ms2pipModel', None)


        # Accession Groups
        self.accession_hierarchy = config_dict.get('accessionKeys')
        self.accession_flags = config_dict.get('accessionFlags')
        self.accession_format = config_dict.get('accessionFormat')
        if self.accession_format == 'invitroSPI' and self.accession_hierarchy is None:
            self.accession_hierarchy = ['nonspliced', 'spliced']

        self.use_accession_stratum = config_dict.get('useAccessionStrata', False)

        # Epitope validation
        self.host_proteome = config_dict.get('hostProteome')
        self.pathogen_proteome = config_dict.get('pathogenProteome')
        self.control_flags = config_dict.get('controlFlags')
        self.alleles = config_dict.get('alleles')
        self.epitope_candidate_cut_off = config_dict.get('epitopeCandidateCutOff', 0.1)
        self.epitope_cut_level = config_dict.get('epitopeCutLevel', 'psm')
        self.host_only_results = config_dict.get('hostOnlyResults')
        self.engine_score_cut = config_dict.get('engineScoreCut')
        self.epitope_length_cut_off = config_dict.get('epitopeLengthCutOff', 15)

    def __str__(self):
        print_string = f'inSPIRE Settings for Experiment {self.experiment_title}:<br>'
        print_string += f"""<table style="width:40%">
        <tr>
            <th>Config</th>
            <th>Setting</th>
        </tr>
        <tr>
            <td>searchEngine</td>
            <td>{self.search_engine}</td>
        </tr>
        <tr>
            <td>scansFormat</td>
            <td>{self.scans_format}</td>
        </tr>
        <tr>
            <td>spectralPredictor</td>
            <td>{self.spectral_predictor}</td>
        </tr>
        <tr>
            <td>deltaMethod</td>
            <td>{self.delta_method}</td>
        </tr>
        <tr>
            <td>rescoreMethod</td>
            <td>{self.rescore_method}</td>
        </tr>
        <tr>
            <td>searchResults</td>
            <td>{self.search_results}</td>
        </tr>
        <tr>
            <td>scansFolder</td>
            <td>{self.scans_folder}</td>
        </tr>
        <tr>
            <td>outputFolder</td>
            <td>{self.output_folder}</td>
        </tr>
        """
        if self.spectral_predictor == 'prosit':
            print_string += f"""
                <tr>
                    <td>collisionEnergy</td>
                    <td>{self.collision_energy}</td>
                </tr>
            """
        else:
            print_string += f"""
                <tr>
                    <td>ms2pipModel</td>
                    <td>{self.ms2pip_model}</td>
                </tr>
            """
        print_string += f"""
            <tr>
                <td>mzAccuracy</td>
                <td>{self.mz_accuracy}</td>
            </tr>
            <tr>
                <td>mzUnits</td>
                <td>{self.mz_units}</td>
            </tr>
            <tr>
                <td>fixedModifications</td>
                <td>{self.fixed_modifications}</td>
            </tr>
            <tr>
                <td>forceReload</td>
                <td>{self.force_reload}</td>
            </tr>
            <tr>
                <td>falseDiscoveryRate</td>
                <td>{self.fdr}</td>
            </tr>
            <tr>
                <td>excludeFeatures</td>
                <td>{self.exclude_features}</td>
            </tr>
            <tr>
                <td>includeFeatures</td>
                <td>{self.include_features}</td>
            </tr>
            <tr>
                <td>reduce</td>
                <td>{self.reduce}</td>
            </tr>
            <tr>
                <td>filterCysteine</td>
                <td>{self.filter_c}</td>
            </tr>
            <tr>
                <td>dropUnknownPTMs</td>
                <td>{self.drop_unknown_mods}</td>
            </tr>
            <tr>
                <td>useBindingAffinity</td>
                <td>{self.use_binding_affinity}</td>
            </tr>
        """

        print_string += '</table>'

        return print_string

    def validate(self):
        """ Function to validate config settings.
        """
        if self.search_engine not in ('mascot', 'maxquant', 'peaks', 'msfragger'):
            raise ValueError(
                f'Unsupported Search Engine: "{self.search_engine}". Supported ' +
                'engines are "mascot", "peaks", "msfragger", and "maxquant".'
            )

        if self.scans_format not in ('mgf', 'mzML'):
            raise ValueError(
                f'Unsupported Scans Format: "{self.scans_format}". Supported ' +
                'formats are "mgf" and "mzML".'
            )

        if self.mz_units not in ('Da', 'ppm'):
            raise ValueError(
                f'Unsupported mz unit: "{self.mz_units}". Supported units are :Da" and "ppm".'
            )

        if self.rescore_method not in ('mokapot', 'percolator', 'percolatorSeparate'):
            raise ValueError(
                f'Unsupported Rescore Method: "{self.rescore_method}". Supported ' +
                'methods are "mokapot", "percolatorSeparate", and "percolator".'
            )

        if self.spectral_predictor not in ('prosit', 'ms2pip'):
            raise ValueError(
                f'Unsupported Spectral Predictor: "{self.spectral_predictor}". Supported ' +
                'methods are "prosit" and "ms2pip".'
            )

        if self.spectral_predictor == 'ms2pip' and self.ms2pip_model is None:
            raise ValueError(
                'You must specify an ms2pipModel when using the ms2pip spectral predictor.'
            )

        if self.epitope_cut_level not in ('psm', 'peptide'):
            raise ValueError(
                'Unsupported epitope filtering level {self.epitope_cut_level}, must be' +
                'psm or peptide.'
            )



""" Functions for generating the html report at the end of the quantification
    pipeline.
"""
import os
import webbrowser

import pandas as pd

from inspire.constants import ENDC_TEXT, OKCYAN_TEXT

def create_quant_report(config):
    """ Function to create the final html report and open it in the brower.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the whole pipeline.
    figures : dict
        A dictionary containing all of the required plots.
    """
    out_path = os.path.abspath(config.output_folder)

    base_table = """<table style="width:40%">
        <tr>
            <th>Original Name</th>
            <th>Renamed</th>
            <th>Sample</th>
        </tr>
    """
    meta_df = pd.read_csv(
        f'{config.output_folder}/quant/metadata.csv'
    )
    if 'sample' not in meta_df.columns:
        meta_df['sample'] = meta_df['renamed'].str.replace('File', 'Sample')

    for _, df_row in meta_df.iterrows():
        base_table += f'''
            <tr>
                <td>{df_row['source']}</td>
                <td>{df_row['renamed']}</td>
                <td>{df_row['sample']}</td>
            </tr>
        '''

    with open(f'{out_path}/img/norm_correlation.svg', mode='r', encoding='UTF-8') as in_f:
        norm_corr = in_f.read()

    with open(f'{out_path}/img/raw_correlation.svg', mode='r', encoding='UTF-8') as in_f:
        raw_corr = in_f.read()


    with open(f'{out_path}/img/quant_pca.svg', mode='r', encoding='UTF-8') as in_f:
        quant_pca = in_f.read()

    with open(f'{out_path}/img/quant_distro.svg', mode='r', encoding='UTF-8') as in_f:
        quant_distro = in_f.read()


    base_table += '</table>'

    html_string = ('''
    <html>
        <head>
            <link 
                rel="stylesheet"
                href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.1/css/bootstrap.min.css"
            >
            <style>
                body{
                    font-family: Helvetica;
                    margin:0 100;
                    padding-bottom: 50px;
                    background:whitesmoke;
                }
                h2{
                    color: firebrick;
                    font-family: Helvetica;
                }
                h3{
                    color: firebrick;
                    font-family: Helvetica;
                }
                h2:after
                {
                    content:' ';
                    display: block;
                    border:0.25px solid #696969;
                    position: absolute;
                    width: 60%;
                    margin-top: 2px;
                    left: 20%;
                }
                table {
                    font-family: Helvetica;
                    width: 60%;
                    border: 2px solid #696969;
                }
                th, td {
                    border: 1px solid #696969;
                    padding: 2px;
                }
            </style>
        </head>
        <body>
            <center>
            <h2>Quantification Report for ''' + config.experiment_title + '''</h2>
            </center>
            <h3>
                Raw Files Analysed
            </h3>
            <p>
                ''' + base_table + '''
            </p>
            <h3>
                Correlation between Files:
            </h3>
            <p>
                The figure shows the correlation between raw intensities. You should see particularly
                high correlation if you have technical replicates.
            </p>
        ''' + raw_corr +
        '''
                <br><br>
                <p>
                    The figure shows the correlation between normalised intensities. As with the
                    non-normalised intensities you should see highest correlation between
                    technical replicates. You may also see some increase in correlation across
                    files after normalisation.
                </p>
        ''' + norm_corr +
        '''
                <h3>
                    Distributions across files:
                </h3>
                <p>
                    This figure shows the distributions of abundances as quantified
                    by Skyline before and after normalisation. While the raw intensities
                    may be higher or lower for individual files, the normalised
                    intensities should all be centered around the same point.
                </p>
        ''' + quant_distro
    )

        
    if (
        os.path.exists(f'{out_path}/img/quant_logo_plots.svg') and
        os.path.exists(f'{out_path}/img/quant_cluster_violin.svg')
    ):
        with open(f'{out_path}/img/quant_logo_plots.svg', mode='r', encoding='UTF-8') as in_f:
            quant_logo = in_f.read()
        with open(f'{out_path}/img/quant_cluster_violin.svg', mode='r', encoding='UTF-8') as in_f:
            quant_clust_violin = in_f.read()

        html_string += (
            '''
                <br><br>
                <h3>Clustering based on Fold Change</h3>
                <p>
                    For this analysis we cluster quantified peptides based on the log2 fold change
                    in their abundance in order to understand the quantitative differences in the
                    immunopeptidome brought about by pathogen infection. Clusters 1, 2, and 3 are
                    peptides under represented after infection, while clusters 5, 6, and 7 are
                    peptides over represented.
                </p>
            ''' + quant_clust_violin +
            '''
                <br><br>
                <p>
                    The figure shows amino acids over and under represented at C-terminus among clusters
                    compared to the central cluster (cluster 4). Considering changes at the C-terminus could
                    be relevant for identification of alternate cleavage mechanisms or changes in HLA-I molecule
                    preference given that C-terminus is an anchor site.
                </p>
            ''' + quant_logo
        )

    html_string += (
        '''
                <h3>
                    Clustering Over Files
                </h3>
                <p>
                    The figure the raw files after principal component analysis of normalised
                    peptide intensities. We should see technical replicates clustering
                    together. Furthermore, we hope to see separation between infected and
                    control files in the case of PEPSeek.
                </p>
        ''' + quant_pca
    )

    if os.path.exists(f'{out_path}/img/quant_clustermap.svg'):
        with open(f'{out_path}/img/quant_clustermap.svg', mode='r', encoding='UTF-8') as in_f:
            quant_clustermap = in_f.read()

        html_string += (
            '''
                <br><br>
                <p>
                    The figure shows a clustering heatmap over normalised intensities.
                </p>
            ''' + quant_clustermap
        )

    if os.path.exists(f'{out_path}/img/peptide_volcano.svg'):
        with open(f'{out_path}/img/peptide_volcano.svg', mode='r', encoding='UTF-8') as in_f:
            pep_volcano = in_f.read()
        html_string += (
            '''
                    <br><br>
                    <h3>
                        Volcano Plots of Up and Downregulated Peptides
                    </h3>
                    <p>
                        This plot shows up and downregulation on the peptide level. These results are
                        also available in csv format in the quantitative outputs.
                    </p>
            ''' + pep_volcano
        )

    output_path = f'{config.output_folder}/quant/quant-report.html'
    with open(output_path, 'w', encoding='UTF-8') as output_file:
        output_file.write(html_string)

    print(
        OKCYAN_TEXT +
        '\tReport generated.' +
        ENDC_TEXT
    )

    if not config.silent_execution:
        webbrowser.open(
            'file://' + os.path.realpath(output_path),
            new=2
        )



""" Functions for RAW file conversion.
"""
from multiprocessing import Pool
import os
from pathlib import Path
import platform

from inspire.constants import ENDC_TEXT, OKCYAN_TEXT
from inspire.download import download_thermo_raw_file_parser

def convert_raw_to_mgf(config):
    """ Function to convert RAW files by calling the ThermoRawFileParser.

    Parameters
    ----------
    config : inspire.config.Config
        Config object which runs the experiment.
    """
    scans_folder = config.scans_folder
    raw_files = [
        scan_f for scan_f in os.listdir(scans_folder) if scan_f.lower().endswith('.raw')
    ]
    print(
        OKCYAN_TEXT + f'\tFound {len(raw_files)} RAW files to be converted.' + ENDC_TEXT
    )

    home = str(Path.home())
    if not os.path.isdir(f'{home}/inSPIRE_models/ThermoRawFileParser'):
        download_thermo_raw_file_parser()
    thermo_path = f'{home}/inSPIRE_models/ThermoRawFileParser/ThermoRawFileParser.exe'

    prefix = ''
    if platform.system() != 'Windows':
        prefix = 'mono '

    convert_commands = [
        f'{prefix}{thermo_path} -o={scans_folder} -f=0 -i={scans_folder}/{raw_file} -l 4'
        for raw_file in raw_files if not os.path.exists(
            f'{scans_folder}/{raw_file.replace(".raw", ".mgf")}'
        )
    ]

    with Pool(processes=config.n_cores) as pool:
        pool.map(os.system, convert_commands)



""" Functions for reading NetMHCpan output.
"""
import os

import pandas as pd
import polars as pl

MHC_PAN_COL_NAMES = [
    'Pos',
    'MHC',
    'Peptide',
    'Core',
    'Of',
    'Gp',
    'Gl',
    'Ip',
    'Il',
    'Icore',
    'Identity',
    'Score_EL',
    '%Rank_EL',
    'Score_BA',
    '%Rank_BA',
    'Aff(nM)',
    'BindLevel',
]


def _get_mhc_pan_metadata(mhc_pan_file):
    """ Function to find where to read from the mhc pan output.

    Parameters
    ----------
    mhc_pan_file : str
        Path to an output of NetMHCpan predicted binding affinity.

    Returns
    -------
    start_line : int
        The line on which tabular data starts.
    end_line : int
        The line on which tabular data ends.
    """
    with open(mhc_pan_file, 'r', encoding='UTF-8') as open_file:
        line_idx = 0
        dash_lines = []

        while len(dash_lines) < 3 and (line := open_file.readline()):
            if '--------' in line:
                dash_lines.append(line_idx)
            line_idx += 1

        if len(dash_lines) < 3:
            return -1, -1
        start_line = dash_lines[1] + 1
        end_line = dash_lines[2] - 1
    return start_line, end_line

def _clean_mhc_pan_output(mhc_pan_file):
    """ Function to deal with inconsistent white spacing in the mhcpan output.

    Parameters
    ----------
    mhc_pan_file : str
        Path to an output of NetMHCpan predicted binding affinity.
    """
    with open(mhc_pan_file, 'r', encoding='UTF-8') as file :
        filedata = file.read()

    filedata = filedata.replace('<= SB', '<=SB')
    filedata = filedata.replace('<= WB', '<=WB')

    with open(mhc_pan_file, 'w', encoding='UTF-8') as file:
        file.write(filedata)

def read_single_mhcpan_file(mhc_pan_file):
    """ Function to read in from a single NetMHCpan output file.

    Parameters
    ----------
    mhc_pan_file : str
        Path to an output of NetMHCpan predicted binding affinities.

    Returns
    -------
    mhc_pan_df : pd.DataFrame
        Tabular data from the file read into a DataFrame.
    """
    _clean_mhc_pan_output(mhc_pan_file)
    start, end = _get_mhc_pan_metadata(mhc_pan_file)
    try:
        mhc_pan_df = pd.read_csv(
            mhc_pan_file,
            delim_whitespace=True,
            skiprows=lambda idx, sl=start, el=end : idx < sl or idx > el,
            header=None,
            names=MHC_PAN_COL_NAMES
        )
    except pd.errors.ParserError:
        mhc_pan_df = pd.read_csv(
            mhc_pan_file,
            delim_whitespace=True,
            skiprows=lambda idx, sl=start, el=end : idx < sl or idx > el,
            header=None,
            names=[col for col in MHC_PAN_COL_NAMES if col != 'BindLevel']
        )

    mhc_pan_df['BindLevel'] = mhc_pan_df['%Rank_BA'].apply(
        lambda x : '<=SB' if x < 0.5 else ('<=WB' if x < 2.0 else '')
    )
    return mhc_pan_df

def read_mhcpan_output(mhc_pan_dir, per_allele=False):
    """ Function read raw mhcpan output files.

    Parameters
    ----------
    mhc_pan_dir : str
        A directory containing NetMHCpan output files.
    per_allele : bool (default=False)
        A flag indicating whether the data should be separated per allele.

    Returns
    -------
    mhc_pan_combined_df : pl.DataFrame
        DataFrame of NetMHCpan predicted binding affinities.
    """
    mhc_pan_files = [
        f'{mhc_pan_dir}/{x}' for x in os.listdir(mhc_pan_dir) if x.startswith('output')
    ]
    mhc_dfs = {
        x.split('_')[-3] + x.split('_')[-1]: [] for x in mhc_pan_files
    }
    for mhc_pan_file in mhc_pan_files:
        input_code = mhc_pan_file.split('_')[-3] + mhc_pan_file.split('_')[-1]
        mhc_pan_df = read_single_mhcpan_file(mhc_pan_file)

        if per_allele:
            allele = mhc_pan_file.split('_')[-2].split('.txt')[0]
            mhc_pan_df = mhc_pan_df.rename(
                columns={
                    'Aff(nM)': f'{allele}_BindingAffinity',
                    '%Rank_BA': f'{allele}_%Rank_BA',
                    'BindLevel': f'{allele}_BindLevel',
                }
            )
            mhc_pan_df = mhc_pan_df[[
                'Peptide',
                f'{allele}_BindingAffinity',
                f'{allele}_%Rank_BA',
                f'{allele}_BindLevel',
            ]]
        mhc_dfs[input_code].append(mhc_pan_df)

    if per_allele:
        all_dfs = []
        for input_code in mhc_dfs:
            per_input_dfs = mhc_dfs[input_code]
            mhc_pan_combined_df = per_input_dfs[0].drop_duplicates(subset=['Peptide'])
            for mhc_df in per_input_dfs[1:]:
                mhc_pan_combined_df = pd.merge(
                    mhc_pan_combined_df,
                    mhc_df.drop_duplicates(subset=['Peptide']),
                    how='inner',
                    on='Peptide',
                )
            all_dfs.append(mhc_pan_combined_df)
    else:
        all_dfs = []
        for length in mhc_dfs:
            all_dfs.extend(mhc_dfs[length])

    mhc_pan_combined_df = pd.concat(all_dfs)
    if not per_allele:
        # If different HLAs have been predicted find the minimum prediction.
        mhc_pan_combined_df['minAffinity'] = mhc_pan_combined_df.groupby(
            ['Peptide']
        )['Aff(nM)'].transform(min)
        mhc_pan_combined_df = mhc_pan_combined_df[
            mhc_pan_combined_df['Aff(nM)'] == mhc_pan_combined_df['minAffinity']
        ]
        mhc_pan_combined_df = mhc_pan_combined_df.drop_duplicates('Peptide')

    return pl.from_pandas(mhc_pan_combined_df)



""" Functions for simple pipeline to get spectral angle on of identified PSMs.
"""
import pandas as pd
import polars as pl

from inspire.constants import(
    CHARGE_KEY,
    KNOWN_PTM_WEIGHTS,
    SCAN_KEY,
    SOURCE_KEY,
    INTENSITIES_KEY,
    MZS_KEY,
    PEPTIDE_KEY,
    PROSIT_INTES_KEY,
    PROSIT_IONS_KEY,
    PROSIT_SEQ_KEY,
    PTM_SEQ_KEY,
    RT_KEY,
)
from inspire.input.msp import msp_to_df
from inspire.input.ssl import ssl_file_to_inspire_format
from inspire.predict_spectra import predict_spectra
from inspire.spectral_features import calculate_spectral_features
from inspire.utils import convert_mod_seq_to_ptm_seq, fetch_scan_data, filter_for_prosit

def get_spectral_angle(config):
    """ Function to generate pair plots of selected PSMs (experimental vs. Prosit
        predicted spectra.).

    Parameters
    ----------
    config : inspire.config.Config
        The Config object used to run the inSPIRE experiment.
    """
    for dataset in config.sa_query_dfs:
        if dataset.endswith('.ssl'):
            input_df = ssl_file_to_inspire_format(dataset)
        else:
            input_df = pl.read_csv(dataset)

        output_loc = dataset.split('.csv')[0] + '_spectralAngle.csv'
        input_cols = list(input_df.columns)

        get_charge_from_scan_file = not CHARGE_KEY in input_df.columns
        get_rt = not RT_KEY in input_df.columns
        scan_df = fetch_scan_data(input_df, config, get_charge_from_scan_file, with_rt=get_rt)

        input_df = input_df.join(
            scan_df,
            how='inner',
            on=[SOURCE_KEY, SCAN_KEY]
        )

        input_df = filter_for_prosit(input_df)

        input_df = input_df.with_columns(
            pl.col('modifiedSequence').apply(
                lambda x : x.replace('[+16.0]', '(ox)').replace('[+57.0]', '')
            ).alias('modified_sequence'),
            pl.col(CHARGE_KEY).alias('precursor_charge'),
        )

        if 'collisionEnergy' in input_df.columns:
            input_df = input_df.rename({'collisionEnergy': 'collision_energy'})
        else:
            input_df = input_df.with_columns(
                pl.lit(config.collision_energy).alias('collision_energy')
            )

        input_df[['modified_sequence', 'precursor_charge', 'collision_energy']].write_csv(
            f'{config.output_folder}/saInput.csv'
        )

        predict_spectra(config, pipeline='spectralAngle')
        prosit_df = msp_to_df(f'{config.output_folder}/saPredictions.msp', 'prosit', None)
        prosit_df = prosit_df.unique(subset=['modified_sequence', CHARGE_KEY])

        input_df = input_df.join(
            prosit_df,
            how='inner',
            on=['modified_sequence', CHARGE_KEY]
        )

        input_df = input_df.with_columns(
            pl.col('modifiedSequence').apply(
                convert_mod_seq_to_ptm_seq
            ).alias('ptm_seq'),
        )

        input_df = input_df.with_columns(
            pl.struct([
                CHARGE_KEY,
                'collisionEnergy',
                INTENSITIES_KEY,
                MZS_KEY,
                PEPTIDE_KEY,
                PROSIT_INTES_KEY,
                PROSIT_IONS_KEY,
                PROSIT_SEQ_KEY,
                PTM_SEQ_KEY,
            ]).apply(lambda df_row : calculate_spectral_features(
                df_row,
                {
                    0: 0.0,
                    1: KNOWN_PTM_WEIGHTS['Oxidation (M)'],
                    2: KNOWN_PTM_WEIGHTS['Carbamidomethylation'],
                },
                config.mz_accuracy,
                config.mz_units,
                None,
                '1',
                config.delta_method,
                minimal_features=True,
            )).alias('spectralResults')
        )
        input_df = input_df.unnest('spectralResults')
        output_columns = ['spectralAngle', 'spearmanR']

        try:
            sources = input_df['source'].unique().tolist()
            for source in sources:
                source_df = input_df[input_df['source'] == source]
                source_rt_df = pd.read_csv(f'{config.output_folder}/rt_fit_{source}.csv')
                intercept = source_rt_df['intercepts'].mean()
                coefficient = source_rt_df['coefficents'].mean()
                source_df['deltaRT'] = source_df[['source', 'retentionTime', 'iRT']].apply(
                    lambda df_row, coef=coefficient, interc=intercept : abs(
                        (df_row['retentionTime'] - (df_row['iRT']*coef)+interc)/coef
                    ),
                    axis=1,
                )
                output_columns.append('deltaRT')
        except Exception as e:
            print(f'Retention Time Prediction Comparison failed with error {e}')

        input_df.select(input_cols + output_columns).write_csv(
            output_loc
        )



""" Function for generating Prosit or MS2PIP spectral predictions.
"""
from pathlib import Path

import numpy as np
import pandas as pd

from inspire.prosit import (
    load_model,
    get_sequence_integer,
    get_precursor_charge_onehot,
    prosit_predict,
    write_msp_file,
)


def predict_spectra(config, pipeline='core'):
    """ Function to generate spectral predictions for a set of peptide sequences
        using either Prosit or MS2PIP and write to msp format.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object which manages the experiment.
    pipeline : str (default=core)
        The pipeline which is being run.
    """
    if config.spectral_predictor == 'ms2pip':
        raise ValueError('inSPIRE 2.0 does not support running MS2PIP natively.')

    home = str(Path.home())
    d_irt = load_model(
        f'{home}/inSPIRE_models/models/irt_config.yml',
        f'{home}/inSPIRE_models/models/irt_model.json',
        f'{home}/inSPIRE_models/models/weight_66_0.00796.hdf5',
    )
    d_spectra = load_model(
        f'{home}/inSPIRE_models/models/spectra_config.yml',
        f'{home}/inSPIRE_models/models/spectra_model.json',
        f'{home}/inSPIRE_models/models/weight_163_0.11385.hdf5',
    )

    if pipeline =='core':
        input_file = f'{config.output_folder}/prositInput.csv'
        out_file = f'{config.output_folder}/prositPredictions.msp'
    elif pipeline == 'calibrate':
        input_file = f'{config.output_folder}/calibrationInput.csv'
        out_file = f'{config.output_folder}/calibrationPredictions.msp'
    elif pipeline == 'spectralAngle':
        input_file = f'{config.output_folder}/saInput.csv'
        out_file = f'{config.output_folder}/saPredictions.msp'
    elif pipeline == 'validation':
        input_file = f'{config.output_folder}/validationInput.csv'
        out_file = f'{config.output_folder}/validationPredictions.msp'
    else:
        input_file = f'{config.output_folder}/plotInput.csv'
        out_file = f'{config.output_folder}/plotPredictions.msp'

    for chunk_idx, input_df in enumerate(pd.read_csv(input_file, chunksize=200_000)):
        input_df = input_df.reset_index(drop=True)
        prosit_input = {
            'collision_energy_aligned_normed': (
                np.expand_dims(
                    np.array(input_df['collision_energy']).astype(float), axis=1
                ) / 100.0
            ),
            'sequence_integer': get_sequence_integer(input_df['modified_sequence']),
            'precursor_charge_onehot': get_precursor_charge_onehot(
                input_df['precursor_charge']
            ),
        }

        prosit_data = prosit_predict(prosit_input, d_irt)
        final_result = prosit_predict(prosit_data, d_spectra)

        write_msp_file(input_df, final_result, out_file, chunk_idx)
        del prosit_input
        del prosit_data



""" Function for dealing with different accession groups.
"""
from copy import deepcopy

import polars as pl

from inspire.constants import (
    ACCESSION_STRATUM_KEY,
    ACCESSION_KEY,
    ENGINE_SCORE_KEY,
    LABEL_KEY,
    PEPTIDE_KEY,
    SOURCE_KEY,
    SCAN_KEY
)
from inspire.utils import fetch_proteome

ACCESSION_SPLITTERS = {
    'mascot': ',',	
    'msfragger': ',',	
    'maxquant': ';',	
    'peaks': ':',	
}

def get_accession_group(accession, search_engine, accession_hierarchy, accession_groups):
    """ Function to get the accession group from the accession of a peptide.

    Parameters
    ----------
    accession : str
        The accession of the peptide.
    search_engine : str
        The search engine which provided the data.
    accession_hierarchy : list of str
        The order in which groups should be assigned in case of multi-mappers.
    accession_groups : dict
        A dictionary of mapping accession groups to the flag that will be seen
        in the accession.
    """
    splitter = ACCESSION_SPLITTERS.get(search_engine)
    if splitter is None:
        raise ValueError(f'Unrecognised Search Engine: {search_engine}')

    if accession in ('deNovo', 'unknown'):
        return len(accession_hierarchy)

    all_possible_accessions = accession.split(splitter)
    assignment = None
    for individiual_accession in all_possible_accessions:
        found = False
        for idx, acc in enumerate(accession_hierarchy):
            if accession_groups[acc] in individiual_accession:
                if assignment is None or idx < assignment:
                    assignment = idx
                found = True
        if not found:
            return 0
    return assignment

def get_invitro_spi_acc_group(accession, search_engine):
    """ Function to get the accession group of a PSM from the accession column in invitroSPI
        format.

    Parameters
    ----------
    accession : str
        The accession of a PSM.
    search_engine : str
        The search engine from which the results were generated.

    Returns
    -------
    assignment : int
        The accession group index as defined by the config.accession_hierarchy.
    """
    splitter = ACCESSION_SPLITTERS.get(search_engine)
    if splitter is None:
        raise ValueError(f'Unrecognised Search Engine: {search_engine}')

    all_possible_accessions = accession.split(splitter)
    assignment = 0
    for individiual_accession in all_possible_accessions:
        if 'PCP_' in individiual_accession:
            return 0
        if 'PSP_' in individiual_accession:
            assignment = 1

    return assignment

def validate_accession_stratum(df_row, proteome, rev_proteome, config):
    """ Function to validate the accession of a non-canonical PSM by checking for the
        sequence in the standard proteome.

    Parameters
    ----------
    df_row : pd.Series
        An entry in a DataFrame of search results.
    proteome : list of tuple
        The standard proteome.
    rev_proteome : list of tuple
        The reversed proteome.
    config : inspire.config.Config
        The Config object for the experiment.

    Returns
    -------
    df_row : pd.Series
        The input row updated with accession and accession group validated.
    """
    pep_seq = df_row[PEPTIDE_KEY].replace('I', 'L')
    if df_row[ACCESSION_STRATUM_KEY] == 0:
        return {
            ACCESSION_KEY: df_row[ACCESSION_KEY],
            ACCESSION_STRATUM_KEY: df_row[ACCESSION_STRATUM_KEY],
        }

    results ={}
    if df_row[LABEL_KEY] == 1:
        for entry in proteome:
            if pep_seq in entry[1]:
                if config.accession_format == 'invitroSPI':
                    position = entry[1].index(pep_seq) + 1
                    results[ACCESSION_KEY] = f'PCP_{position}_{position+len(pep_seq)}'
                    results[ACCESSION_STRATUM_KEY] = 0
                else:
                    results[ACCESSION_KEY] = entry[0]
                    results[ACCESSION_STRATUM_KEY] = 0
    else:
        for entry in rev_proteome:
            if pep_seq in entry[1]:
                if config.accession_format == 'invitroSPI':
                    position = entry[1].index(pep_seq) + 1
                    results[ACCESSION_KEY] = f'PCP_{position}_{position+len(pep_seq)}'
                    results[ACCESSION_STRATUM_KEY] = 0
                else:
                    results[ACCESSION_KEY] = entry[0]
                    results[ACCESSION_STRATUM_KEY] = 0

    return results


def process_accession_groups(main_df, config):
    """ Function to process the accession groups of a DataFrame of search results.

    Parameters
    ----------
    main_df : pd.DataFrame
        The DataFrame of search results.
    config : inspire.config.Config
        The Config object for the experiment.

    Returns
    -------
    main_df : pd.DataFrame
        The input DataFrame updated with accession group column, validated accession column
        and removed any PSMs of accession groups to ignore.
    """
    if config.accession_format == 'invitroSPI':
        main_df = main_df.with_columns(pl.col(ACCESSION_KEY).apply(
            lambda x : get_invitro_spi_acc_group(x, config.search_engine)
        ).alias(ACCESSION_STRATUM_KEY))
    else:
        main_df = main_df.with_columns(
            pl.col(ACCESSION_KEY).apply(
                lambda x : get_accession_group(
                    x,
                    config.search_engine,
                    config.accession_hierarchy,
                    config.accession_flags
                ) if isinstance(x, str) else 0
            ).alias(ACCESSION_STRATUM_KEY)
        )

    if config.proteome is not None:
        prot_sequences = fetch_proteome(config.proteome, with_desc=False)

        rev_prot_seqs = []
        for entry in deepcopy(prot_sequences):
            new_entry = ('reversed_' + entry[0], entry[1][::-1])
            rev_prot_seqs.append(new_entry)

        main_df = main_df.with_columns(
            pl.struct([ACCESSION_KEY, ACCESSION_STRATUM_KEY, PEPTIDE_KEY, LABEL_KEY]).apply(
                lambda x : validate_accession_stratum(x, prot_sequences, rev_prot_seqs, config),
            ).alias('results')
        )
        main_df = main_df.drop([ACCESSION_KEY, ACCESSION_STRATUM_KEY])
        main_df = main_df.unnest('results')

    main_df = main_df.sort(
        by=[LABEL_KEY, ACCESSION_STRATUM_KEY, ENGINE_SCORE_KEY, PEPTIDE_KEY],
        descending=[True, False, True, False],
    )
    main_df = main_df.unique(subset=[SOURCE_KEY, SCAN_KEY, PEPTIDE_KEY])

    if 'ignore' in config.accession_hierarchy:
        ignore_idx = config.accession_hierarchy.index('ignore')
        main_df = main_df.filter(pl.col(ACCESSION_STRATUM_KEY).ne(ignore_idx))

    return main_df
